/*-------------------------------------------------------------------------+
| Copyright 2020 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.editor.fx.controller;

import static java.util.Objects.requireNonNull;
import static org.fortiss.tooling.base.ui.utils.LayoutDataUIUtils.getConnectionPoints;
import static org.fortiss.tooling.base.ui.utils.LayoutDataUIUtils.removeConnectionPoint;

import java.util.List;

import org.eclipse.emf.common.util.EList;
import org.fortiss.tooling.base.model.layout.ILayoutedModelElement;
import org.fortiss.tooling.base.model.layout.Point;
import org.fortiss.tooling.common.ui.javafx.lwfxef.DiagramViewer;
import org.fortiss.tooling.common.ui.javafx.lwfxef.controller.IController;
import org.fortiss.tooling.common.ui.javafx.lwfxef.controller.base.LinkControllerBase;
import org.fortiss.tooling.common.ui.javafx.lwfxef.model.IModelChangeProvider;
import org.fortiss.tooling.common.ui.javafx.lwfxef.mvc.ILinkMVCBundle;
import org.fortiss.tooling.common.ui.javafx.lwfxef.mvc.IMVCBundle;
import org.fortiss.tooling.kernel.model.INamedCommentedElement;
import org.fortiss.tooling.kernel.service.IConnectionCompositorService;

/**
 * Base class for {@link IController}s for {@link ILayoutedModelElement}s representing links and
 * connections in a {@link DiagramViewer}.
 * 
 * @author munaro
 */
public abstract class LayoutedLinkBendPointController<T extends INamedCommentedElement & ILayoutedModelElement>
		extends LinkControllerBase {
	/** {@link IModelChangeProvider} for this controller */
	protected final NamedCommentedLayoutModelChangeProvider<T> layoutModelChangeProvider;

	/** Returns the model element. */
	@SuppressWarnings("unchecked")
	public T getModelElement() {
		return (T)getModel();
	}

	/** {@inheritDoc} */
	@Override
	public IModelChangeProvider getModelChangeProvider() {
		return layoutModelChangeProvider;
	}

	/** Constructor. */
	public LayoutedLinkBendPointController(ILinkMVCBundle mvcb, Class<T> modelType) {
		super(mvcb);

		Object model = requireNonNull(mvcb.getModel(), "The given model is null!");
		if(!modelType.isAssignableFrom(model.getClass())) {
			throw new IllegalArgumentException("Expected model of type " +
					modelType.getSimpleName() + ", but was " + model.getClass().getSimpleName());
		}

		@SuppressWarnings("unchecked") T lme = (T)mvcb.getModel();
		this.layoutModelChangeProvider = new NamedCommentedLayoutModelChangeProvider<T>(lme);
	}

	/** {@inheritDoc} */
	@Override
	public void delete() {
		deleteLink();
	}

	/** {@inheritDoc} */
	@Override
	protected boolean clampAndSnapBendPointToGridCenter() {
		// disable default mode
		return false;
	}

	/** {@inheritDoc} */
	@Override
	protected boolean clampAndSnapBendPointToGrid() {
		// enable clamping mode
		return true;
	}

	/** {@inheritDoc} */
	@Override
	protected void deleteBendPoint(int bpIndex) {
		removeConnectionPoint(getModelElement(), bpIndex);
	}

	/** {@inheritDoc} */
	@Override
	protected void updateModelAfterBendPointMove(int bpIndex, double dx, double dy) {
		List<Point> oldPoints = getBendPointList(getModelElement());
		Point point = oldPoints.get(bpIndex);
		int oldX = point.getX();
		int oldY = point.getY();
		int newX = oldX + (int)dx;
		int newY = oldY + (int)dy;
		point.setX(newX);
		point.setY(newY);
	}

	/** {@inheritDoc} */
	@Override
	protected void deleteLink() {
		IConnectionCompositorService.getInstance().disconnect(getModelElement());
	}

	/** {@inheritDoc} */
	@Override
	protected void updateModelAfterLinkMove(ILinkMVCBundle linkBundle, IMVCBundle startBundle,
			IMVCBundle endBundle) {
		// ignored
	}

	/** Returns the list of bend-points. */
	protected EList<Point> getBendPointList(ILayoutedModelElement modelElement) {
		return getConnectionPoints(modelElement).getPoints();
	}

	/** {@inheritDoc} */
	@Override
	protected int getNumerOfBendPoints() {
		return getBendPointList(getModelElement()).size();
	}
}
