/*-------------------------------------------------------------------------+
| Copyright 2019 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.editor.fx.controller;

import static java.util.Objects.requireNonNull;
import static org.fortiss.tooling.base.ui.utils.LWFXEditorUtils.convertSideToEOrientation;
import static org.fortiss.tooling.base.utils.LayoutDataUtils.getNodeSize;
import static org.fortiss.tooling.base.utils.LayoutDataUtils.moveNode;
import static org.fortiss.tooling.base.utils.LayoutDataUtils.setNodeSize;
import static org.fortiss.tooling.base.utils.LayoutDataUtils.setStickyConnectorLayoutData;

import org.fortiss.tooling.base.model.element.IConnector;
import org.fortiss.tooling.base.model.layout.Dimension;
import org.fortiss.tooling.base.model.layout.ILayoutedModelElement;
import org.fortiss.tooling.common.ui.javafx.lwfxef.FeedbackChange;
import org.fortiss.tooling.common.ui.javafx.lwfxef.controller.IController;
import org.fortiss.tooling.common.ui.javafx.lwfxef.mvc.IContentAnchorageMVCBundle;
import org.fortiss.tooling.common.ui.javafx.lwfxef.mvc.IContentMVCBundle;
import org.fortiss.tooling.kernel.model.INamedCommentedElement;

import javafx.geometry.Dimension2D;
import javafx.geometry.Side;

/**
 * {@link IController} for {@link ILayoutedModelElement}s that have a rectangular shape.
 * 
 * @author diewald
 */
public class LayoutedRectangularResizableContentController<T extends INamedCommentedElement & ILayoutedModelElement>
		extends EObjectRectangularResizableContentControllerBase<T> {
	/** Constructor. */
	public LayoutedRectangularResizableContentController(IContentMVCBundle mvcb,
			Class<T> modelType) {
		super(mvcb, modelType);

		Object model = requireNonNull(mvcb.getModel(), "The given model is null!");
		if(!modelType.isAssignableFrom(model.getClass())) {
			throw new IllegalArgumentException("Expected model of type " +
					modelType.getSimpleName() + ", but was " + model.getClass().getSimpleName());
		}
	}

	/** {@inheritDoc} */
	@Override
	protected void moveAnchorageToSideOffset(IContentAnchorageMVCBundle anchorage, Side side,
			double offset) {
		Object connObj = anchorage.getModel();
		if(connObj instanceof IConnector && connObj instanceof ILayoutedModelElement) {
			ILayoutedModelElement conn = (ILayoutedModelElement)connObj;

			Dimension2D portDim = anchorage.getVisual().getDimensions();
			int correction;
			if(side == Side.TOP || side == Side.BOTTOM) {
				correction = (int)portDim.getWidth() / 2;
			} else {
				correction = (int)portDim.getHeight() / 2;
			}
			int corrected = (int)offset + correction;
			setStickyConnectorLayoutData(conn, convertSideToEOrientation(side), corrected);
		}
	}

	/** {@inheritDoc} */
	@Override
	protected void move(FeedbackChange deltaChange) {
		moveNode(getModelElement(), (int)deltaChange.getDeltaX(), (int)deltaChange.getDeltaY());
	}

	/** {@inheritDoc} */
	@Override
	protected void resize(FeedbackChange delta) {
		T element = getModelElement();
		Dimension d = getNodeSize(element);
		int w = d.getWidth() + (int)delta.getDeltaW();
		int h = d.getHeight() + (int)delta.getDeltaH();
		setNodeSize(element, w, h);
	}
}
