/*-------------------------------------------------------------------------+
| Copyright 2019 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.editor.fx.controller;

import static org.fortiss.tooling.base.model.layout.LayoutPackage.ILAYOUTED_MODEL_ELEMENT__LAYOUT_DATA;

import org.eclipse.emf.common.notify.Adapter;
import org.eclipse.emf.common.notify.Notification;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.util.EContentAdapter;
import org.fortiss.tooling.base.model.layout.ILayoutData;
import org.fortiss.tooling.base.model.layout.ILayoutedModelElement;
import org.fortiss.tooling.base.model.layout.LayoutPackage;
import org.fortiss.tooling.kernel.model.FortissToolingKernelPackage;
import org.fortiss.tooling.kernel.model.INamedCommentedElement;

/**
 * A {@link KernelServiceBasedModelChangeProviderBase} for {@link ILayoutedModelElement}s.
 * 
 * @author hoelzl
 */
class NamedCommentedLayoutModelChangeProvider<T extends INamedCommentedElement & ILayoutedModelElement>
		extends KernelServiceBasedModelChangeProviderBase<T> {
	/** The layouted model element. */
	private final T modelElement;
	/** The layout EMF notification listener. */
	private final LayoutModelElementAdapter notificationListener;

	/** Constructor. */
	public NamedCommentedLayoutModelChangeProvider(T modelElement) {
		this.modelElement = modelElement;
		this.notificationListener = new LayoutModelElementAdapter();
	}

	/** {@inheritDoc} */
	@Override
	protected T getModelElement() {
		return modelElement;
	}

	/** {@inheritDoc} */
	@Override
	protected Adapter getNotificationListener() {
		return notificationListener;
	}

	/**
	 * Checks whether the given notification should be accepted.
	 * 
	 * @param notification
	 *            the notification
	 * @return whether the notification should be accepted
	 */
	protected boolean acceptNotification(Notification notification) {
		return false;
	}

	/** Specific {@link EContentAdapter} listening to layout data changes. */
	public class LayoutModelElementAdapter extends EContentAdapter {
		/** {@inheritDoc} */
		@Override
		public void notifyChanged(Notification notification) {
			if(acceptNotification(notification)) {
				addNotification(notification);
				return;
			}
			Object notifier = notification.getNotifier();
			if(notifier == modelElement) {
				Object feature = notification.getFeature();
				if(feature == FortissToolingKernelPackage.Literals.INAMED_ELEMENT__NAME ||
						feature == FortissToolingKernelPackage.Literals.INAMED_COMMENTED_ELEMENT__COMMENT) {
					addNotification(notification);
					return;
				} else if(feature == LayoutPackage.Literals.ILAYOUTED_MODEL_ELEMENT__LAYOUT_DATA) {
					// self adapt to layout data
					super.notifyChanged(notification);
					addNotification(notification);
					return;
				}
			} else if(notifier instanceof ILayoutData) {
				// self adapt to layout data
				super.notifyChanged(notification);
				addNotification(notification);
				return;
			}
			// prevent self-adapt for all other cases
		}

		/** Called when layout data of the underlying model element changed. */
		protected void layoutChanged() {
			// default does nothing
		}

		/** {@inheritDoc} */
		@Override
		protected void selfAdapt(Notification notification) {
			// ignore resources and resource sets
			if(notification.getNotifier() instanceof EObject) {
				super.selfAdapt(notification);
			}
		}

		/** {@inheritDoc} */
		@Override
		protected void handleContainment(Notification notification) {
			// only adapt to layout data
			if(notification.getNotifier() instanceof ILayoutData) {
				super.handleContainment(notification);
				return;
			}

			Class<?> instanceClass =
					LayoutPackage.Literals.ILAYOUTED_MODEL_ELEMENT.getInstanceClass();
			// only adapt to layout data
			if(notification.getFeatureID(instanceClass) == ILAYOUTED_MODEL_ELEMENT__LAYOUT_DATA) {
				super.handleContainment(notification);
			}
		}
	}
}
