/*-------------------------------------------------------------------------+
| Copyright 2019 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.editor.fx.visual;

import static java.util.Collections.emptyList;
import static org.fortiss.tooling.base.ui.utils.LWFXEditorUtils.computeLinkToCircleLocation;
import static org.fortiss.tooling.base.ui.utils.LayoutDataUIUtils.getConnectionPoints;

import java.util.List;

import org.fortiss.tooling.common.ui.javafx.lwfxef.DiagramCoordinate;
import org.fortiss.tooling.common.ui.javafx.lwfxef.mvc.ILinkMVCBundle;
import org.fortiss.tooling.common.ui.javafx.lwfxef.visual.rectangular.LineLinkVisualBase;
import org.fortiss.tooling.base.model.layout.ILayoutedModelElement;
import org.fortiss.tooling.base.model.layout.Point;
import org.fortiss.tooling.base.model.layout.Points;

import javafx.geometry.Rectangle2D;

/**
 * {@link LineLinkVisualBase} providing a default visual for wires in standard
 * boxes-and-wires diagrams based on {@link ILayoutedModelElement}s.
 * 
 * @author munaro
 */
public class LayoutedLineLinkVisual<T extends ILayoutedModelElement> extends LineLinkVisualBase {

	/** Constructor. */
	public LayoutedLineLinkVisual(ILinkMVCBundle mvcb, Class<T> modelType) {
		super(mvcb);

		// TODO(#3877): Move type checks to a common base class.
		Object model = mvcb.getModel();
		if(model == null) {
			throw new IllegalArgumentException("The given model is null!");
		}
		if(!modelType.isAssignableFrom(model.getClass())) {
			throw new IllegalArgumentException("Expected model of type " +
					modelType.getSimpleName() + ", but was " + model.getClass().getSimpleName());
		}
	}

	/** Returns the model element with the expected type. */
	@SuppressWarnings("unchecked")
	protected T getModelElement() {
		return (T)getModel();
	}

	/** {@inheritDoc} */
	@Override
	public DiagramCoordinate getBendPointLocation(int bpIndex) {
		Point point = getBendPointList().get(bpIndex);
		return new DiagramCoordinate(point.getX(), point.getY());
	}

	/** {@inheritDoc} */
	@Override
	protected DiagramCoordinate getStartAnchorLocation(Rectangle2D anchorBounds,
			DiagramCoordinate target) {
		return computeLinkToCircleLocation(anchorBounds, target);
	}

	/** {@inheritDoc} */
	@Override
	protected DiagramCoordinate getEndAnchorLocation(Rectangle2D anchorBounds,
			DiagramCoordinate target) {
		return computeLinkToCircleLocation(anchorBounds, target);
	}

	/** {@inheritDoc} */
	@Override
	protected Object getBendPointModel(int i) {
		return getBendPointList().get(i);
	}

	/** {@inheritDoc} */
	@Override
	protected int getNumberOfBendPoints() {
		return getBendPointList().size();
	}

	/** Returns the list of bend-points. */
	private List<Point> getBendPointList() {
		Points connectionPoints = getConnectionPoints(getModelElement());
		if(connectionPoints == null) {
			return emptyList();
		}
		List<Point> pointList = connectionPoints.getPoints();
		if(pointList == null) {
			return emptyList();
		}
		return pointList;
	}

	/** {@inheritDoc} */
	@Override
	protected boolean showArrowOnLastSegment() {
		return true;
	}

	/** {@inheritDoc} */
	@Override
	protected double getInvisibleSelectionLineWidth() {
		return 3;
	}
}
