/*-------------------------------------------------------------------------+
| Copyright 2011 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.utils;

import static org.fortiss.tooling.base.layout.DefaultLayoutConstants.DEFAULT_CONNECTOR_SIZE;
import static org.fortiss.tooling.base.layout.DefaultLayoutConstants.DEFAULT_SHAPE_INSETS;
import static org.fortiss.tooling.base.layout.DefaultLayoutConstants.DEFAULT_SHAPE_MINIMUM_HEIGHT;
import static org.fortiss.tooling.base.layout.DefaultLayoutConstants.DEFAULT_SHAPE_MINIMUM_WIDTH;
import static org.fortiss.tooling.base.ui.utils.LayoutDataUIUtils.getAbsolute;
import static org.fortiss.tooling.base.ui.utils.LayoutDataUIUtils.getConnectorOffsetOrientation;
import static org.fortiss.tooling.base.utils.LayoutDataUtils.getNodePosition;
import static org.fortiss.tooling.base.utils.LayoutDataUtils.getNodeSize;

import org.eclipse.draw2d.geometry.Dimension;
import org.eclipse.draw2d.geometry.Point;
import org.eclipse.draw2d.geometry.Rectangle;
import org.fortiss.tooling.base.model.layout.ILayoutedModelElement;
import org.fortiss.tooling.base.model.layout.OffsetOrientation;
import org.fortiss.tooling.base.utils.LayoutModelElementFactory;
import org.fortiss.tooling.base.utils.RectangleLayoutUtils;

/**
 * Utility class for rectangle-shaped edit part layout.
 * 
 * @author wandinger
 * @author hummel
 * @author hoelzl
 */
public class RectangleLayoutUIUtils {

	/**
	 * Returns the {@link OffsetOrientation} of the point on the rectangle
	 * border having a given Point inside the rectangle.
	 * 
	 * @param source
	 *            the relative coordinates in the rectangle
	 * @param rectangleSize
	 *            the size of the rectangle
	 * @param inset
	 *            the inset from each side
	 * @return the offset and orientation
	 */
	public static OffsetOrientation calculateRectangleBorderOrientation(Point source,
			Dimension rectangleSize, int inset) {
		org.fortiss.tooling.base.model.layout.Point modelPoint = LayoutModelElementFactory
				.createPoint(source.x(), source.y(), source.toString() + "_element");
		org.fortiss.tooling.base.model.layout.Dimension modelDimension =
				LayoutModelElementFactory.createDimension(rectangleSize.width(),
						rectangleSize.height(), rectangleSize.toString() + "_model");
		return RectangleLayoutUtils.calculateRectangleBorderOrientation(modelPoint, modelDimension,
				inset);
	}

	/**
	 * Ensure that the minimum size of a rectangular shape is redeemed.
	 * 
	 * @param oldSize
	 *            the old size to be adjusted
	 * @param minimumWidth
	 *            the minimum to be redeemed
	 * @param minimumHeight
	 *            the minimum height to be redeemed
	 * @return the new dimensions
	 */
	public static Dimension clampRectangularShapeSize(Dimension oldSize, int minimumWidth,
			int minimumHeight) {
		// check for minimum size
		Dimension newSize =
				new Dimension(oldSize.width < minimumWidth ? minimumWidth : oldSize.width,
						oldSize.height < minimumHeight ? minimumHeight : oldSize.height);
		return newSize;
	}

	/**
	 * Ensure that the minimum size of a rectangular shape is redeemed.
	 * 
	 * @param oldSize
	 *            the old size to be adjusted
	 * @return the new dimensions
	 */
	public static Dimension retainMinimumRectangularShapeSize(Dimension oldSize) {
		return clampRectangularShapeSize(oldSize, DEFAULT_SHAPE_MINIMUM_WIDTH,
				DEFAULT_SHAPE_MINIMUM_HEIGHT);
	}

	/**
	 * Returns the bounds rectangle of the connector given its parent bounds.
	 * 
	 * @param connector
	 *            the connector with relative positions
	 * @param parent
	 *            the parent container
	 * @returns the absolute rectangle
	 */
	public static Rectangle getAbsoluteConnectorBounds(ILayoutedModelElement connector,
			ILayoutedModelElement parent) {
		Rectangle result = new Rectangle(0, 0, DEFAULT_CONNECTOR_SIZE, DEFAULT_CONNECTOR_SIZE);
		// get orientation
		OffsetOrientation orient = getConnectorOffsetOrientation(connector);
		// get parent size
		org.fortiss.tooling.base.model.layout.Dimension parentSize = getNodeSize(parent);
		org.fortiss.tooling.base.model.layout.Point pos =
				getAbsolute(orient, parentSize, DEFAULT_SHAPE_INSETS);
		org.fortiss.tooling.base.model.layout.Point parentPos = getNodePosition(parent);
		int maxX = parentSize.getWidth() - DEFAULT_CONNECTOR_SIZE;
		int maxY = parentSize.getHeight() - DEFAULT_CONNECTOR_SIZE;
		result.x = parentPos.getX() + ((pos.getX() > maxX) ? maxX : pos.getX());
		result.y = parentPos.getY() + ((pos.getY() > maxY) ? maxY : pos.getY());
		return result;
	}
}
