/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.annotation;

import static org.fortiss.tooling.kernel.utils.EcoreUtils.pickFirstInstanceOf;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Queue;
import java.util.Set;

import org.eclipse.emf.ecore.EObject;
import org.fortiss.tooling.base.annotation.valueprovider.IAnnotationValueProvider;
import org.fortiss.tooling.base.model.element.IAnnotatedSpecification;
import org.fortiss.tooling.base.model.element.IModelElement;
import org.fortiss.tooling.kernel.ToolingKernelActivator;
import org.fortiss.tooling.kernel.extension.data.ITopLevelElement;
import org.fortiss.tooling.kernel.service.IKernelIntrospectionSystemService;
import org.fortiss.tooling.kernel.service.IPersistencyService;
import org.fortiss.tooling.kernel.service.base.EObjectAwareServiceBase;
import org.fortiss.tooling.kernel.service.types.EMFTypeMap;
import org.fortiss.tooling.kernel.utils.EcoreUtils;
import org.fortiss.tooling.kernel.utils.LoggingUtils;

/**
 * Concrete implementation of service used to retrieve and instantiate model element annotations.
 * It evaluates the <code>org.fortiss.tooling.base.annotation</code> Eclipse extension point to
 * obtain the information which annotations types should be bound to which model elements.
 * 
 * @author eder
 */
public class AnnotationValueService
		extends EObjectAwareServiceBase<IAnnotationValueProvider<IAnnotatedSpecification>>
		implements IAnnotationValueService {
	/** Singleton instance of the service. */
	public static final AnnotationValueService INSTANCE = new AnnotationValueService();

	/** Returns the service instance. */
	public static AnnotationValueService getInstance() {
		return INSTANCE;
	}

	/** The compositor extension point ID. */
	private static final String EXTENSION_POINT_NAME = "org.fortiss.tooling.base.annotation";

	/** The compositor configuration element name. */
	private static final String CONFIGURATION_ELEMENT_NAME = "annotation";

	/** The handler class attribute name. */
	private static final String HANDLER_CLASS_ATTRIBUTE_NAME = "binding";

	/** Cache for lookup of registered value providers. */
	private EMFTypeMap<EObject, List<IAnnotationValueProvider<IAnnotatedSpecification>>> valueProviderCache =
			new EMFTypeMap<>();

	/** Cache to lookup {@link AnnotationEntry}s for {@link IModelElement}s. */
	private Map<IModelElement, AnnotationEntry> annotationEntryCache =
			new HashMap<IModelElement, AnnotationEntry>();

	/** {@inheritDoc} */
	@Override
	public void startService() {
		IKernelIntrospectionSystemService.getInstance().registerService(this);
	}

	/** {@inheritDoc} */
	@Override
	public String getIntrospectionDescription() {
		return "The Annotation Value Service contributes model element attributes that always exist if the corresponding model element is instantiated.";
	}

	/** {@inheritDoc} */
	@Override
	public String getIntrospectionLabel() {
		return "Annotation Value Service";
	}

	/**
	 * <p>
	 * In contrast to the implementation of
	 * {@link EObjectAwareServiceBase#getRegisteredHandlers(Class)}, this implementation of
	 * {@link #getRegisteredHandlers(Class)} returns all {@link IAnnotationValueProvider}s that have
	 * been registered for the given class or any of its super-classes or interfaces.
	 * </p>
	 */
	@SuppressWarnings("unchecked")
	@Override
	public List<IAnnotationValueProvider<IAnnotatedSpecification>>
			getRegisteredHandlers(Class<?> modelElementClass) {

		// Check if there is a cached result
		List<IAnnotationValueProvider<IAnnotatedSpecification>> valueProviderList =
				valueProviderCache.get(modelElementClass);
		if(valueProviderList != null) {
			return valueProviderList;
		}

		Set<IAnnotationValueProvider<IAnnotatedSpecification>> valueProviderSet =
				new HashSet<IAnnotationValueProvider<IAnnotatedSpecification>>();

		// Queue to walk super-type hierarchy
		Queue<Class<?>> q = new LinkedList<Class<?>>();
		q.add(modelElementClass);

		while(!q.isEmpty()) {
			Class<?> current = q.poll();

			// Determine registered IAnnotationValueProviders for the current class
			// (nearest match)
			List<IAnnotationValueProvider<IAnnotatedSpecification>> currentHandlers =
					super.getRegisteredHandlers(current);
			if(currentHandlers != null) {
				valueProviderSet.addAll(currentHandlers);
			}

			// Also, investigate super-types
			Class<?> superClass = current.getSuperclass();
			if(superClass != null && superClass != Object.class) {
				q.add(superClass);
			}

			for(Class<?> iface : current.getInterfaces()) {
				q.add(iface);
			}
		}

		valueProviderList =
				new ArrayList<IAnnotationValueProvider<IAnnotatedSpecification>>(valueProviderSet);

		// Cache the result as the lookup that walks the entire super-type hierarchy is really
		// expensive (in particular since super.getRegisteredHandlers(current) also partially walks
		// the super-type hierarchy to perform a nearest match.
		valueProviderCache.put((Class<? extends EObject>)modelElementClass, valueProviderList);

		return valueProviderList;
	}

	/**
	 * Prepares the {@link AnnotationEntry} for a given {@link IModelElement} and ensures that all
	 * {@link IAnnotatedSpecification}s are instantiated.
	 */
	private void prepareAnnotationEntry(final IModelElement element, AnnotationEntry entry) {
		final List<IAnnotationValueProvider<IAnnotatedSpecification>> registeredHandlers =
				getRegisteredHandlers(element.getClass());

		for(IAnnotationValueProvider<IAnnotatedSpecification> annotationProvider : registeredHandlers) {

			IAnnotatedSpecification annotatedSpecification = EcoreUtils.pickFirstInstanceOf(
					annotationProvider.getAnnotationClazz(), element.getSpecifications());

			// TODO (see #2228) Keep this code that ensures that non-existing annotations are
			// instantiated when the respective model elements are displayed in an annotation view
			// until #2228 is resolved. While ElementCompositorService ensures that annotations are
			// instantiated when models are composed interactively, this code is still required for
			// models that are created programmatically.

			// It should be noted that each time an annotation is constructed from this
			// context, an undo/redo command will be created that is not comprehensible for the user
			// (#2208)
			if(annotatedSpecification == null) {
				// TODO (see #2228): Emit a warning when annotations are missing. Until the
				// method, AnnotationUtils.getAnnotation(...) is changed to run without being
				// wrapped into command. Currently, the warning has a performance impact on
				// programmatic model transformation.

				// String elementName =
				// element instanceof INamedElement ? computeFullyQualifiedName(
				// (INamedElement)element, true) : element.toString();
				// LoggingUtils
				// .warning(
				// ToolingKernelActivator.getDefault(),
				// "Annotation \"" +
				// annotationProvider.getAnnotationClazz() +
				// "\" is missing for model element \"" +
				// elementName +
				// "\". If this model element has been created programmatically, ensure that the
				// \"IElementCompositorService\" is used to link it to the model (which ensures the
				// instantiation of annotations).");
				try {
					annotatedSpecification =
							annotationProvider.getAnnotatedSpecificationForModelElement(element);
				} catch(Exception ex) {
					LoggingUtils.error(ToolingKernelActivator.getDefault(),
							"Error instantating new annotation: createAnnotatedSpecificationForModelElement() failed for " +
									annotationProvider.getAnnotationClazz() + ".",
							ex);
					return;
				}
			}

			// Register annotation provider <-> annotation specification mapping
			entry.addNewSpecification(annotationProvider, annotatedSpecification);
		}
	}

	/** {@inheritDoc} */
	@Override
	public void instantiateAnnotations(final IModelElement element) {
		getAnnotationEntry(element, false);
	}

	/** {@inheritDoc} */
	@Override
	public boolean requiresAnnotationInstantiation(final IModelElement element) {
		final List<IAnnotationValueProvider<IAnnotatedSpecification>> registeredHandlers =
				getRegisteredHandlers(element.getClass());
		if(registeredHandlers == null) {
			return false;
		}

		for(IAnnotationValueProvider<IAnnotatedSpecification> annotationProvider : registeredHandlers) {
			IAnnotatedSpecification annotatedSpecification = pickFirstInstanceOf(
					annotationProvider.getAnnotationClazz(), element.getSpecifications());

			if(annotatedSpecification == null) {
				return true;
			}
		}
		return false;
	}

	/**
	 * Helper method that performs the actual work of {@link #getAnnotationEntry(IModelElement)} and
	 * {@link #instantiateAnnotations(IModelElement)}, i.e. that acquires (and potentially
	 * instantiates) annotations for model elements.
	 * 
	 * @return {@link AnnotationEntry} for all {@link IAnnotatedSpecification}s of the given
	 *         {@link IModelElement}.
	 */
	private AnnotationEntry getAnnotationEntry(final IModelElement element,
			boolean invokedFromGuiContext) {

		AnnotationEntry cachedResult = annotationEntryCache.get(element);
		if(cachedResult != null) {
			return cachedResult;
		}

		if(element == null) {
			return null;
		}

		final AnnotationEntry result = new AnnotationEntry(element);

		// Wrap instantiation of new annotation into command if this method is invoked from the
		// GUI context, i.e. from getAnnotationEntry(final IModelElement element), but only if
		// prepareAnnotationEntry() would actually modify the underlying model, and a model
		// context is available.
		final ITopLevelElement modelContext =
				IPersistencyService.getInstance().getTopLevelElementFor(element);
		if(requiresAnnotationInstantiation(element) && modelContext != null) {
			modelContext.runAsCommand(new Runnable() {

				@Override
				public void run() {
					prepareAnnotationEntry(element, result);
				}
			});
		} else {
			prepareAnnotationEntry(element, result);
		}

		// Only put element into AnnotationEntry <-> IModelElement cache if method is invoked from
		// the GUI context, i.e. from getAnnotationEntry(final IModelElement element).
		//
		// This avoids polluting the cache with entries for temporary models (created during a
		// model-transformation that invoked instantiateAnnotations(final IModelElement element).
		//
		// In case instantiateAnnotations(final IModelElement element) is invoked from
		// HierarchicElementCompositorBase, the cache will be updated when the respective model
		// element is displayed for the first time.
		if(invokedFromGuiContext) {
			annotationEntryCache.put(element, result);
		}

		return result;
	}

	/** {@inheritDoc} */
	@Override
	public AnnotationEntry getAnnotationEntry(final IModelElement element) {
		return getAnnotationEntry(element, true);
	}

	/** {@inheritDoc} */
	@Override
	protected String getExtensionPointName() {
		return EXTENSION_POINT_NAME;
	}

	/** {@inheritDoc} */
	@Override
	protected String getConfigurationElementName() {
		return CONFIGURATION_ELEMENT_NAME;
	}

	/** {@inheritDoc} */
	@Override
	protected String getHandlerClassAttribute() {
		return HANDLER_CLASS_ATTRIBUTE_NAME;
	}

	/** {@inheritDoc} */
	@Override
	protected boolean requiresUniqueHandler() {
		return false;
	}
}
