/*******************************************************************************
 * Copyright (c) 2016, 2018 fortiss GmbH.
 * 
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v2.0 which is available at
 * http://www.eclipse.org/legal/epl-v20.html
 * 
 * SPDX-License-Identifier: EPL-2.0
 * 
 * Contributors:
 *     Florian Hoelzl (fortiss GmbH) - initial implementation
 *     Alexander Diewald (fortiss GmbH) - backport to AF3 & Integration
 *                                        improvements
 *
 *******************************************************************************/
package org.fortiss.tooling.common.ui.javafx.control.treetableview;

import static org.apache.commons.lang3.SystemUtils.IS_OS_LINUX;

import javafx.beans.value.ChangeListener;
import javafx.scene.control.ContextMenu;
import javafx.scene.control.TreeCell;
import javafx.scene.control.TreeItem;
import javafx.scene.control.TreeView;
import javafx.scene.input.ClipboardContent;
import javafx.scene.input.DragEvent;
import javafx.scene.input.Dragboard;
import javafx.scene.input.MouseButton;
import javafx.scene.input.MouseEvent;
import javafx.scene.input.TransferMode;

/**
 * A JavaFX {@link TreeView} based on a {@link DynamicTreeContentProviderBase} and a
 * {@link DynamicTreeUIProviderBase}. This class is intended to be used as a wrapper for
 * {@link TreeView} and its rather complex expert API.
 */
public final class DynamicTreeViewer<T> extends DynamicTreeViewerBase<T> {
	/** The {@link TreeView} control to be managed. */
	private final TreeView<T> view;
	/** {@link TreeItem} constructed for the root object. */
	private final DynamicTreeItem<T> rootItem;
	/** The UI provider implementation. */
	private final DynamicTreeUIProviderBase<T> uiProvider;
	/** The selection change listener. */
	private ChangeListener<Object> selectionChangeListener = null;

	/** Constructor. */
	public DynamicTreeViewer(TreeView<T> view, T root, boolean showRoot, int revealLevel,
			DynamicTreeContentProviderBase<T> contentProvider,
			DynamicTreeUIProviderBase<T> uiProvider) {
		super(view, contentProvider);
		this.uiProvider = uiProvider;
		// construct view
		this.view = view;
		rootItem = new DynamicTreeItem<T>(root, contentProvider);
		view.setRoot(rootItem);
		view.setShowRoot(showRoot);
		configureCellFactory();
		view.getSelectionModel().selectedItemProperty().addListener((obs, oldVal, newVal) -> {
			if(selectionChangeListener != null) {
				Object oVal = oldVal != null ? oldVal.getValue() : null;
				Object nVal = newVal != null ? newVal.getValue() : null;
				selectionChangeListener.changed(obs, oVal, nVal);
			}
		});
		rootItem.update();
		// expand to reveal (+1 if root node is not shown)
		expandItem(rootItem, showRoot ? revealLevel : revealLevel + 1);
	}

	/** Constructor. */
	public DynamicTreeViewer(T root, boolean showRoot, int revealLevel,
			DynamicTreeContentProviderBase<T> contentProvider,
			DynamicTreeUIProviderBase<T> uiProvider) {
		this(new TreeView<T>(), root, showRoot, revealLevel, contentProvider, uiProvider);
	}

	/** Updates the tree view. */
	public void update() {
		// wild cast works: see constructor
		DynamicTreeItem<T> rootItem = (DynamicTreeItem<T>)view.getRoot();
		rootItem.update();
	}

	/** Sets the selection of this tree view. */
	public void setSelection(Object value) {
		// wild cast works: see constructor
		DynamicTreeItem<T> rootItem = (DynamicTreeItem<T>)view.getRoot();
		DynamicTreeItem<T> item = findItem(rootItem, value);
		int row = view.getRow(item);
		view.getSelectionModel().select(row);
	}

	/** Searches the {@link DynamicTreeItem} for the given value from the root. */
	public DynamicTreeItem<T> findItem(T value) {
		return findItem(rootItem, value);
	}

	/** Searches the {@link DynamicTreeItem} for the given value. */
	public DynamicTreeItem<T> findItem(DynamicTreeItem<T> item, Object value) {
		if(item == null || value == null || item.getValue() == null) {
			return null;
		}
		if(item.getValue().equals(value)) {
			return item;
		}
		for(TreeItem<T> subItem : item.getChildren()) {
			if(subItem instanceof DynamicTreeItem) {
				DynamicTreeItem<T> i = findItem((DynamicTreeItem<T>)subItem, value);
				if(i != null) {
					return i;
				}
			}
		}
		return null;
	}

	/** Sets the selection change listener. */
	public void setSelectionListener(ChangeListener<Object> listener) {
		this.selectionChangeListener = listener;
	}

	/** Configures the cell factory for the viewer. */
	private void configureCellFactory() {
		view.setCellFactory(param -> {
			TreeCell<T> cell = new TreeCell<T>() {
				@Override
				public void updateItem(T item, boolean empty) {
					super.updateItem(item, empty);
					ContextMenu menu = null;
					if(!empty && item != null) {
						menu = uiProvider.createContextMenu(item);
						if(IS_OS_LINUX && menu != null) {
							// JFX-builtin handling is non-operation in Linux.
							addContextMenuHandler(menu, item);
						}

						this.setText(uiProvider.getLabel(item));
						this.setGraphic(uiProvider.getIconNode(item));
						this.setOnDragDetected(evt -> {
							dragDetected(evt, this, item);
						});
						this.setOnDragOver(evt -> {
							dragOver(evt, item);
						});
						this.setOnDragDropped(evt -> {
							dragDropped(evt, item);
						});
					} else {
						this.setText(null);
						this.setGraphic(null);
					}
					this.setContextMenu(menu);
				}
			};
			return cell;
		});
	}

	/**
	 * Adds a handler for showing the context menu. Duplicate of the same method in
	 * {@link DynamicTreeTableUIProviderBase} due to the absence of proper selection interfaces in
	 * JFX.
	 */
	private void addContextMenuHandler(ContextMenu menu, T element) {
		view.addEventHandler(MouseEvent.MOUSE_RELEASED, e -> {
			if(e.getButton() == MouseButton.SECONDARY) {
				TreeItem<T> selected = view.getSelectionModel().getSelectedItem();

				// We must check the element since all rows are evaluated.
				if(selected != null && selected.getValue() == element) {
					menu.show(view, e.getScreenX(), e.getScreenY());
					e.consume();
				}
			} else {
				// any other click cause hiding menu
				menu.hide();
			}
		});
	}

	/** Called when an item in the tree is dragged. */
	private void dragDetected(MouseEvent evt, TreeCell<T> cell, T item) {
		ClipboardContent cbContent = uiProvider.getDragClipboardContent(item);
		if(cbContent != null) {
			Dragboard db = cell.startDragAndDrop(TransferMode.ANY);
			db.setContent(cbContent);
			evt.consume();
		}
	}

	/** Gives the current drag-over event to the UI provider. */
	private void dragOver(DragEvent evt, T item) {
		if(uiProvider.dragOver(item, evt.getDragboard())) {
			evt.consume();
		}
	}

	/** Gives the completed drop gesture to the UI provider. */
	private void dragDropped(DragEvent evt, T item) {
		uiProvider.dropClipboardContent(item, evt.getDragboard());
	}

	/** Expands the tree to the given item. */
	public void expandItem(TreeItem<T> item) {
		while(item.getParent() != null) {
			item = item.getParent();
			item.setExpanded(true);
		}
	}

	/** Expands items up to the given level. */
	private void expandItem(TreeItem<T> parentItem, int revealLevel) {
		if(revealLevel <= 0) {
			return;
		}
		parentItem.setExpanded(true);
		revealLevel--;
		for(TreeItem<T> child : parentItem.getChildren()) {
			expandItem(child, revealLevel);
		}
	}

	/** Expands the whole tree. */
	public void expandAllItems() {
		expandAllItems(rootItem);
	}

	/** Recursively expands the whole tree. */
	private void expandAllItems(TreeItem<T> item) {
		item.setExpanded(true);
		for(TreeItem<T> child : item.getChildren()) {
			expandAllItems(child);
		}
	}

	/** Returns the underlying {@link TreeView}. */
	public TreeView<T> getControl() {
		return view;
	}

	/** Returns the UI provider. */
	public DynamicTreeUIProviderBase<T> getUIProvider() {
		return uiProvider;
	}

	/** {@inheritDoc} */
	@Override
	protected Object getSelectedObject() {
		DynamicTreeItem<T> item = (DynamicTreeItem<T>)view.getSelectionModel().getSelectedItem();

		if(item != null) {
			return item.getValue();
		}

		return null;
	}
}
