/*******************************************************************************
 * Copyright (c) 2016, 2018 fortiss GmbH. 
 * 
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v2.0 which is available at
 * http://www.eclipse.org/legal/epl-v20.html
 * 
 * SPDX-License-Identifier: EPL-2.0
 * 
 * Contributors:
 *     Florian Hoelzl (fortiss GmbH) - initial implementation
 *     Alexander Diewald (fortiss GmbH) - backport to AF3 & Integration
 *                                        improvements
 *
 *******************************************************************************/
package org.fortiss.tooling.common.ui.javafx.control.treetableview;

import javafx.event.Event;
import javafx.event.EventDispatchChain;
import javafx.event.EventDispatcher;
import javafx.scene.control.Control;
import javafx.scene.control.TreeItem;
import javafx.scene.input.MouseButton;
import javafx.scene.input.MouseEvent;

/**
 * Base class for {@link DynamicTreeViewer} and {@link DynamicTreeTableViewer}.
 */
abstract class DynamicTreeViewerBase<T> {
	/** The double click handler. */
	private IDoubleClickHandler doubleClickHandler;

	/** The double click event dispatcher. */
	protected DoubleClickEventDispatcher doubleClickEventDispatcher;

	/** The content provider implementation. */
	private final DynamicTreeContentProviderBase<T> contentProvider;

	/** Constructor. */
	public DynamicTreeViewerBase(Control control,
			DynamicTreeContentProviderBase<T> contentProvider) {
		this.contentProvider = contentProvider;

		setupEventDispatcher(control);
	}

	/** Creates and installs the EventDispatcher. Might be overwritten to change behavior. */
	protected void setupEventDispatcher(Control control) {
		EventDispatcher original = control.getEventDispatcher();
		control.setEventDispatcher(new DoubleClickEventDispatcher(original));
	}

	/** Returns the content provider. */
	public final DynamicTreeContentProviderBase<T> getContentProvider() {
		return contentProvider;
	}

	/**
	 * Customized event dispatcher to fix double-click behavior. This prevents opening/closing of
	 * tree items with double click listeners assigned.
	 */
	private class DoubleClickEventDispatcher implements EventDispatcher {
		/** The original event dispatcher. */
		private final EventDispatcher original;

		/** Constructor. */
		public DoubleClickEventDispatcher(EventDispatcher original) {
			this.original = original;
		}

		/** {@inheritDoc} */
		@Override
		public Event dispatchEvent(Event event, EventDispatchChain tail) {
			if(event instanceof MouseEvent) {
				MouseEvent mevt = (MouseEvent)event;
				if(mevt.getButton() == MouseButton.PRIMARY && mevt.getClickCount() >= 2) {
					if(mevt.getEventType().equals(MouseEvent.MOUSE_CLICKED) &&
							doubleClickHandler != null) {
						Object item = getSelectedObject();
						if(item != null) {
							doubleClickHandler.doubleClicked(item);
						}
					}
					event.consume();
				}
			}
			return original.dispatchEvent(event, tail);
		}
	}

	/** Sets the double click handler. */
	public void setDoubleClickHandler(IDoubleClickHandler handler) {
		this.doubleClickHandler = handler;
	}

	/**
	 * Retrieves the currently selected {@link Object} from this viewer.
	 * 
	 * Note: This usually is not e.g. the {@link TreeItem} or similar, but the model object, that is
	 * represented by it. Usually this will be some EObject.
	 * 
	 * @return The currently selected {@link Object}.
	 */
	protected abstract Object getSelectedObject();
}
