/*-------------------------------------------------------------------------+
| Copyright 2019 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.common.ui.javafx.layout;

import static java.util.Arrays.asList;
import static javafx.scene.layout.AnchorPane.setBottomAnchor;
import static javafx.scene.layout.AnchorPane.setLeftAnchor;
import static javafx.scene.layout.AnchorPane.setRightAnchor;
import static javafx.scene.layout.AnchorPane.setTopAnchor;
import static org.fortiss.tooling.common.ui.javafx.util.JavaFXUtils.loadFXML;

import java.io.IOException;
import java.net.URL;
import java.util.List;

import javafx.fxml.FXML;
import javafx.scene.Node;
import javafx.scene.layout.AnchorPane;

/**
 * JavaFX node which can be hierarchically composed. It can be included within another
 * {@link CompositeFXControllerBase} and might contain other {@link CompositeFXControllerBase}s.
 * 
 * @param <T>
 *            Type of the controlled element itself.
 * @param <S>
 *            Type of the layout elements which can be contained.
 * 
 * @author munaro
 */
public abstract class CompositeFXControllerBase<T extends Node, S extends Node>
		implements ICompositeFXController<T, S> {

	/**
	 * JavaFX {@link Node} with the layout specified in the {@link FXML} resource and
	 * associated with a controller of type {@code T}.
	 */
	private T layout;

	/**
	 * {@link ICompositeFXController}s to be added to the {@link Node} by the
	 * {@code addContainments} method.
	 */
	private ICompositeFXController<? extends S, ?>[] containments;

	/** Constructor. */
	@SuppressWarnings("unchecked")
	public CompositeFXControllerBase(ICompositeFXController<? extends S, ?>... containments) {
		this.containments = containments;
	}

	/** {@inheritDoc} */
	@Override
	public T getLayout() {
		return layout;
	}

	/** {@inheritDoc} */
	@Override
	public List<ICompositeFXController<? extends S, ?>> getContainments() {
		return asList(containments);
	}

	/** {@inheritDoc} */
	@Override
	public void loadLayout() {
		getContainments().forEach(containment -> containment.loadLayout());
		if(layout == null) {
			URL fxmlResource = getClass().getResource(getFXMLLocation());
			try {
				layout = loadFXML(fxmlResource, this);
			} catch(IOException e) {
				throw new RuntimeException(
						"Cannot load the resource located at " + getFXMLLocation(), e);
			}
		}
	}

	/**
	 * If the container is an {@link AnchorPane}, all margins are removed so as to fit the child to
	 * the anchor pane's size.
	 */
	protected void fitToParent(Node child) {
		if(child.getParent() instanceof AnchorPane) {
			setBottomAnchor(child, 0.0);
			setTopAnchor(child, 0.0);
			setLeftAnchor(child, 0.0);
			setRightAnchor(child, 0.0);
		}
	}
}
