/*******************************************************************************
 * Copyright (c) 2019, 2020 fortiss GmbH. 
 * 
 * This program and the accompanying materials are made available under the
 * terms of the Apache License, Version 2.0 which is available at
 * https://www.apache.org/licenses/LICENSE-2.0.
 *
 * SPDX-License-Identifier: Apache-2.0
 *******************************************************************************/
package org.fortiss.tooling.common.ui.javafx.lwfxef;

import static org.fortiss.tooling.common.ui.javafx.lwfxef.DiagramViewerDefaultTags.FOCUS_TAG;
import static org.fortiss.tooling.common.ui.javafx.lwfxef.DiagramViewerDefaultTags.PRIMARY_SELECTION_TAG;
import static org.fortiss.tooling.common.ui.javafx.lwfxef.DiagramViewerDefaultTags.SECONDARY_SELECTION_TAG;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.function.Consumer;

import org.fortiss.tooling.common.ui.javafx.lwfxef.change.Change;
import org.fortiss.tooling.common.ui.javafx.lwfxef.controller.IController;
import org.fortiss.tooling.common.ui.javafx.lwfxef.controller.IControllerFactory;
import org.fortiss.tooling.common.ui.javafx.lwfxef.model.IModelFactory;
import org.fortiss.tooling.common.ui.javafx.lwfxef.mvc.IAnchorageMVCBundle;
import org.fortiss.tooling.common.ui.javafx.lwfxef.mvc.IContentAnchorageMVCBundle;
import org.fortiss.tooling.common.ui.javafx.lwfxef.mvc.IContentMVCBundle;
import org.fortiss.tooling.common.ui.javafx.lwfxef.mvc.IContentMVCBundleWithParent;
import org.fortiss.tooling.common.ui.javafx.lwfxef.mvc.IDiagramAnchorageMVCBundle;
import org.fortiss.tooling.common.ui.javafx.lwfxef.mvc.IDiagramMVCBundle;
import org.fortiss.tooling.common.ui.javafx.lwfxef.mvc.ILinkMVCBundle;
import org.fortiss.tooling.common.ui.javafx.lwfxef.mvc.IMVCBundle;
import org.fortiss.tooling.common.ui.javafx.lwfxef.mvc.impl.ContentAnchorageMVCBundle;
import org.fortiss.tooling.common.ui.javafx.lwfxef.mvc.impl.ContentMVCBundle;
import org.fortiss.tooling.common.ui.javafx.lwfxef.mvc.impl.DiagramAnchorageMVCBundle;
import org.fortiss.tooling.common.ui.javafx.lwfxef.mvc.impl.DiagramMVCBundle;
import org.fortiss.tooling.common.ui.javafx.lwfxef.mvc.impl.LinkMVCBundle;
import org.fortiss.tooling.common.ui.javafx.lwfxef.mvc.impl.MVCBundleBase;
import org.fortiss.tooling.common.ui.javafx.lwfxef.visual.IContentAnchorageVisual;
import org.fortiss.tooling.common.ui.javafx.lwfxef.visual.IContentVisual;
import org.fortiss.tooling.common.ui.javafx.lwfxef.visual.IDiagramAnchorageVisual;
import org.fortiss.tooling.common.ui.javafx.lwfxef.visual.ILinkVisual;
import org.fortiss.tooling.common.ui.javafx.lwfxef.visual.IVisualFactory;

/**
 * This class encapsulates all methods for handling {@link IMVCBundle}s (including selection)
 * management and is solely used by {@link DiagramViewer} (i.e. its purpose is to reduce the size of
 * the {@link DiagramViewer} class).
 */
/* package */final class MVCBundleManager {
	/** The diagram viewer. */
	private final DiagramViewer viewer;
	/** The diagram layers. */
	private final DiagramLayers layers;
	/** The selection changed callback used to notify about selection changes. */
	private final Consumer<DiagramViewerSelection> selectionChangedCallback;
	/**
	 * The model modifier is called once a change (or a set of changes) needs to be applied to the
	 * underlying data model. If this reference is {@code null}, the framework simply calls the
	 * controller implementations.
	 */
	private final Consumer<Change> modelModifier;
	/** The factory for model elements displayed within this viewer. */
	/* package */ final IModelFactory modelFactory;
	/** The factory for visuals of this viewer. */
	/* package */ final IVisualFactory visualFactory;
	/** The factory for controllers of this viewer. */
	/* package */ final IControllerFactory controllerFactory;
	/** The grid canvas visual. */
	/* package */ GridCanvasVisual gridCanvasVisual;
	/** The {@link IMVCBundle}s for content elements. */
	/* package */ Map<Object, IContentMVCBundle> contentBundles = new HashMap<>();
	/** The {@link IMVCBundle}s for diagram anchorage elements. */
	/* package */ Map<Object, IDiagramAnchorageMVCBundle> diagramAnchorageBundles = new HashMap<>();
	/** The {@link IMVCBundle}s for content anchorage elements. */
	/* package */ Map<Object, IContentAnchorageMVCBundle> contentAnchorageBundles = new HashMap<>();
	/** The {@link IMVCBundle}s for link elements. */
	/* package */ Map<Object, ILinkMVCBundle> linkBundles = new HashMap<>();
	/** The list of possible link targets. */
	/* package */ List<IMVCBundle> possibleLinkTargets = new ArrayList<IMVCBundle>();
	/** The {@link IDiagramMVCBundle} of this viewer. */
	/* package */ IDiagramMVCBundle diagramBundle;
	/** The selected {@link IMVCBundle}. */
	/* package */ IMVCBundle primarySelectedBundle;
	/** The list of secondary selected {@link IMVCBundle}. */
	/* package */ List<IMVCBundle> secondarySelectedBundles = new LinkedList<>();

	/** Constructor. */
	/* package */ MVCBundleManager(DiagramViewer viewer, IModelFactory modelFactory,
			IVisualFactory visualFactory, IControllerFactory controllerFactory,
			DiagramLayers layers, Consumer<DiagramViewerSelection> selectionChangedCallback,
			Consumer<Change> modelModifier) {
		this.viewer = viewer;
		this.gridCanvasVisual = new GridCanvasVisual(viewer);
		this.modelFactory = modelFactory;
		this.visualFactory = visualFactory;
		this.controllerFactory = controllerFactory;
		this.layers = layers;
		this.selectionChangedCallback = selectionChangedCallback;
		this.modelModifier = modelModifier;
	}

	/** Returns the model modifier. */
	public Consumer<Change> getModelModifier() {
		return modelModifier;
	}

	/** Sets the selected {@link IMVCBundle} as single primary selection. */
	/* package */ void setSingleSelectedMVCBundle(IMVCBundle sel) {
		if(primarySelectedBundle == sel) {
			return;
		}
		// single selection on other than primary clears multi-selection
		for(IMVCBundle mvcb : secondarySelectedBundles) {
			mvcb.removeTag(SECONDARY_SELECTION_TAG);
			mvcb.getVisual().updateNodes(layers);
		}
		secondarySelectedBundles.clear();
		// single selection clears old selection
		IMVCBundle oldSelection = primarySelectedBundle;
		if(oldSelection != null) {
			oldSelection.removeTag(FOCUS_TAG);
			oldSelection.removeTag(PRIMARY_SELECTION_TAG);
			oldSelection.getVisual().updateNodes(layers);
		}
		primarySelectedBundle = sel;
		if(primarySelectedBundle != null) {
			primarySelectedBundle.getVisual().requestFocus();
			primarySelectedBundle.addTag(FOCUS_TAG);
			primarySelectedBundle.addTag(PRIMARY_SELECTION_TAG);
			primarySelectedBundle.getVisual().updateNodes(layers);
		}
		fireSelectionChanged();
	}

	/**
	 * Adds the given bundle as primary selection and moves the current selection to the secondary
	 * selections list.
	 */
	/* package */ void addSelectedMVCBundle(IMVCBundle sel) {
		if(primarySelectedBundle == sel || secondarySelectedBundles.contains(sel)) {
			return;
		}
		// make current primary selection into a secondary selection
		if(primarySelectedBundle != null) {
			primarySelectedBundle.removeTag(PRIMARY_SELECTION_TAG);
			primarySelectedBundle.addTag(SECONDARY_SELECTION_TAG);
			secondarySelectedBundles.add(primarySelectedBundle);
			primarySelectedBundle.getVisual().updateNodes(layers);
		}
		// set new primary selection
		primarySelectedBundle = sel;
		if(primarySelectedBundle != null) {
			primarySelectedBundle.addTag(PRIMARY_SELECTION_TAG);
			primarySelectedBundle.getVisual().updateNodes(layers);
		}
		fireSelectionChanged();
	}

	/**
	 * Removes the given bundle from either the primary selection (selecting next the secondary list
	 * element) or from the secondary selection list.
	 */
	/* package */ void removeSelectedMVCBundle(IMVCBundle sel) {
		if(primarySelectedBundle == sel) {
			sel.removeTag(PRIMARY_SELECTION_TAG);
			if(secondarySelectedBundles.isEmpty()) {
				primarySelectedBundle = null;
			} else {
				primarySelectedBundle =
						secondarySelectedBundles.remove(secondarySelectedBundles.size() - 1);
				primarySelectedBundle.removeTag(SECONDARY_SELECTION_TAG);
				primarySelectedBundle.addTag(PRIMARY_SELECTION_TAG);
				primarySelectedBundle.getVisual().updateNodes(layers);
			}
		} else {
			sel.removeTag(SECONDARY_SELECTION_TAG);
			secondarySelectedBundles.remove(sel);
		}
		sel.getVisual().updateNodes(layers);
		fireSelectionChanged();
	}

	/** Calls the {@link #selectionChangedCallback} to inform it about a selection change. */
	private void fireSelectionChanged() {
		if(selectionChangedCallback != null) {
			selectionChangedCallback.accept(getSelection());
		}
	}

	/** Initializes the viewer content. */
	/* package */ void updateContentObjects() {
		possibleLinkTargets.clear();
		// remember old selection
		Object oldSelection =
				primarySelectedBundle != null ? primarySelectedBundle.getModel() : null;
		if(oldSelection == null) {
			setSingleSelectedMVCBundle(null);
		}
		// before creating any content, allow model factory to update
		modelFactory.update();
		// initialize the new hash maps
		Map<Object, IContentMVCBundle> newContentBundles = new HashMap<>();
		Map<Object, IDiagramAnchorageMVCBundle> newDiagramAnchorageBundles = new HashMap<>();
		Map<Object, IContentAnchorageMVCBundle> newContentAnchorageBundles = new HashMap<>();
		Map<Object, ILinkMVCBundle> newLinkBundles = new HashMap<>();
		// construct model by moving MVCBs from the old maps to the new ones
		updateDiagramBundle();
		oldSelection = updateContentAndContentAnchorageBundles(oldSelection, newContentBundles,
				newContentAnchorageBundles);
		oldSelection = updateDiagramAnchorages(oldSelection, newDiagramAnchorageBundles);
		oldSelection = updateLinkBundles(oldSelection, newDiagramAnchorageBundles,
				newContentAnchorageBundles, newLinkBundles);
		// select the diagram if the old selection belongs to a bundle which is not
		// contained in the new bundle maps
		if(oldSelection != null) {
			setSingleSelectedMVCBundle(diagramBundle);
		}
		// iterate through all old maps and remove MVCB references
		detachAndRemoveRemainingMVCB();
		// set new maps active
		this.contentBundles = newContentBundles;
		this.diagramAnchorageBundles = newDiagramAnchorageBundles;
		this.contentAnchorageBundles = newContentAnchorageBundles;
		this.linkBundles = newLinkBundles;
		// update visuals
		updateAllVisuals();
	}

	/** Performs an update of the link bundles. */
	private Object updateLinkBundles(Object oldSelection,
			Map<Object, IDiagramAnchorageMVCBundle> newDiagramAnchorageBundles,
			Map<Object, IContentAnchorageMVCBundle> newContentAnchorageBundles,
			Map<Object, ILinkMVCBundle> newLinkBundles) {
		for(Object link : modelFactory.getLinkModels()) {
			ILinkMVCBundle bundle = linkBundles.remove(link);
			if(bundle == null) {
				bundle = initLinks(link, newDiagramAnchorageBundles, newContentAnchorageBundles);
			} else {
				reinitLinks(bundle, newDiagramAnchorageBundles, newContentAnchorageBundles);
			}
			newLinkBundles.put(link, bundle);
			if(link == oldSelection) {
				setSingleSelectedMVCBundle(bundle);
				oldSelection = null;
			}
		}
		return oldSelection;
	}

	/** Performs an update of the diagram anchorage bundles. */
	private Object updateDiagramAnchorages(Object oldSelection,
			Map<Object, IDiagramAnchorageMVCBundle> newDiagramAnchorageBundles) {
		for(Object m : modelFactory.getDiagramAnchorageModels()) {
			IDiagramAnchorageMVCBundle bundle = diagramAnchorageBundles.remove(m);
			if(bundle == null) {
				bundle = initDiagramAnchorages(m);
			}
			newDiagramAnchorageBundles.put(m, bundle);
			if(m == oldSelection) {
				setSingleSelectedMVCBundle(bundle);
				oldSelection = null;
			}
			if(bundle.getController().isPossibleLinkTarget()) {
				possibleLinkTargets.add(bundle);
			}
		}
		return oldSelection;
	}

	/** Performs an update of the parent relationship of content bundles. */
	private void updateParentRelationship(Map<Object, IContentMVCBundle> newContentBundles,
			List<?> contentModels) {
		for(Object m : contentModels) {
			Object parentModel = modelFactory.getParent(m);
			if(parentModel != null) {
				IContentMVCBundle parentBundle = newContentBundles.get(parentModel);
				IContentMVCBundleWithParent elementBundle =
						(IContentMVCBundleWithParent)newContentBundles.get(m);
				if(elementBundle != null && parentBundle != null) {
					elementBundle.setParentBundle(parentBundle);
				}
			}
		}
	}

	/** Performs an update of the content bundles and content anchorage bundles. */
	private Object updateContentAndContentAnchorageBundles(Object oldSelection,
			Map<Object, IContentMVCBundle> newContentBundles,
			Map<Object, IContentAnchorageMVCBundle> newContentAnchorageBundles) {
		List<?> contentModels = modelFactory.getContentModels();
		for(Object m : contentModels) {
			IContentMVCBundle bundle = contentBundles.remove(m);
			if(bundle == null) {
				bundle = initContentBundle(m);
			}
			newContentBundles.put(m, bundle);
			if(m == oldSelection) {
				setSingleSelectedMVCBundle(bundle);
				oldSelection = null;
			}
			if(bundle.getController().isPossibleLinkTarget()) {
				possibleLinkTargets.add(bundle);
			}
			// initialize content anchorages
			for(Object cElement : modelFactory.getContentAnchorageModels(m)) {
				IContentAnchorageMVCBundle cBundle = contentAnchorageBundles.remove(cElement);
				if(cBundle == null) {
					cBundle = initSingleAttachedContentAnchorage(bundle, cElement);
				}
				newContentAnchorageBundles.put(cElement, cBundle);
				if(cElement == oldSelection) {
					setSingleSelectedMVCBundle(cBundle);
					oldSelection = null;
				}
				if(cBundle.getController().isPossibleLinkTarget()) {
					possibleLinkTargets.add(cBundle);
				}
			}
			// clear child bundles (will be setup again in the next loop if child has not
			// disappeared)
			bundle.getChildren().clear();
		}
		// setup parent relationship between content
		updateParentRelationship(newContentBundles, contentModels);

		return oldSelection;
	}

	/** Performs an update of the diagram bundle. */
	private void updateDiagramBundle() {
		boolean attached = true;
		if(diagramBundle != null && diagramBundle.getModel() != modelFactory.getRootModel()) {
			gridCanvasVisual.removeAllVisuals(layers);
			diagramBundle.detach();
			diagramBundle.remove();
			gridCanvasVisual = null;
			diagramBundle = null;
		}
		if(diagramBundle == null) {
			diagramBundle = new DiagramMVCBundle(modelFactory.getRootModel(), viewer);
			gridCanvasVisual = new GridCanvasVisual(viewer);
			diagramBundle.setVisual(gridCanvasVisual);
			attached = false;
		}
		IController diagramController = controllerFactory.createDiagramController(diagramBundle);
		diagramBundle.setController(diagramController);
		if(!attached) {
			diagramBundle.attach();
		}
		if(diagramController.isPossibleLinkTarget()) {
			possibleLinkTargets.add(diagramBundle);
		}
	}

	/** Detaches and removes all remaining {@link IMVCBundle}s. */
	private void detachAndRemoveRemainingMVCB() {
		for(ILinkMVCBundle b : linkBundles.values()) {
			b.getVisual().removeAllVisuals(layers);
			b.detach();
			b.remove();
		}
		for(IContentAnchorageMVCBundle b : contentAnchorageBundles.values()) {
			b.getVisual().removeAllVisuals(layers);
			b.detach();
			b.remove();
		}
		for(IDiagramAnchorageMVCBundle b : diagramAnchorageBundles.values()) {
			b.getVisual().removeAllVisuals(layers);
			b.detach();
			b.remove();
		}
		for(IContentMVCBundle b : contentBundles.values()) {
			b.getVisual().removeAllVisuals(layers);
			b.detach();
			b.remove();
		}
	}

	/**
	 * Initializes the {@link MVCBundleBase} for the given element and adds it to the
	 * visuals.
	 */
	private IContentMVCBundle initContentBundle(Object m) {
		IContentMVCBundle mvcb = new ContentMVCBundle(m, diagramBundle, viewer);
		IContentVisual visual = visualFactory.createContentVisual(mvcb);
		mvcb.setVisual(visual);
		IController contentController = controllerFactory.createContentController(mvcb);
		mvcb.setController(contentController);
		mvcb.attach();
		return mvcb;
	}

	/**
	 * Initializes the {@link MVCBundleBase} for the given element and adds it to the
	 * visuals.
	 */
	private DiagramAnchorageMVCBundle initDiagramAnchorages(Object m) {
		DiagramAnchorageMVCBundle mvcb = new DiagramAnchorageMVCBundle(m, diagramBundle, viewer);
		IDiagramAnchorageVisual visual = visualFactory.createDiagramAnchorageVisual(mvcb);
		mvcb.setVisual(visual);
		IController diagramAnchorageController =
				controllerFactory.createDiagramAnchorageController(mvcb);
		mvcb.setController(diagramAnchorageController);
		mvcb.attach();
		return mvcb;
	}

	/** Handles the focus reaction when the focus property is set to {@code focus}. */
	/* package */ void handleFocus(boolean focus, DiagramLayers layers) {
		if(primarySelectedBundle == null) {
			return;
		}
		if(focus) {
			primarySelectedBundle.addTag(FOCUS_TAG);
		} else {
			primarySelectedBundle.removeTag(FOCUS_TAG);
		}
		primarySelectedBundle.getVisual().updateNodes(layers);
	}

	/** Adds a content anchorage for the given element and content bundle. */
	private IContentAnchorageMVCBundle initSingleAttachedContentAnchorage(IContentMVCBundle pBundle,
			Object cElement) {
		IContentAnchorageMVCBundle cBundle =
				new ContentAnchorageMVCBundle(cElement, pBundle, viewer);
		IContentAnchorageVisual cVisual = visualFactory.createContentAnchorageVisual(cBundle);
		cBundle.setVisual(cVisual);
		IController contentAnchorageController =
				controllerFactory.createContentAnchorageController(cBundle);
		cBundle.setController(contentAnchorageController);
		cBundle.attach();
		return cBundle;
	}

	/** Initializes the link {@link ILinkMVCBundle}s. */
	private ILinkMVCBundle initLinks(Object link,
			Map<Object, IDiagramAnchorageMVCBundle> diagramAnchorageBundles,
			Map<Object, IContentAnchorageMVCBundle> contentAnchorageBundles) {
		Object startModel = modelFactory.getLinkStart(link);
		if(startModel == null) {
			return null;
		}
		IAnchorageMVCBundle startBundle = getBundleForAnchorageModel(startModel,
				diagramAnchorageBundles, contentAnchorageBundles);
		Object endModel = modelFactory.getLinkEnd(link);
		if(endModel == null) {
			return null;
		}
		IAnchorageMVCBundle endBundle = getBundleForAnchorageModel(endModel,
				diagramAnchorageBundles, contentAnchorageBundles);
		LinkMVCBundle lBundle =
				new LinkMVCBundle(link, diagramBundle, startBundle, endBundle, viewer);
		ILinkVisual visual = visualFactory.createLinkVisual(lBundle);
		lBundle.setVisual(visual);
		IController linkController = controllerFactory.createLinkController(lBundle);
		lBundle.setController(linkController);
		lBundle.attach();
		if(linkController.isPossibleLinkTarget()) {
			possibleLinkTargets.add(lBundle);
		}
		return lBundle;
	}

	/** Updates source and target of the link {@link ILinkMVCBundle}s. */
	private void reinitLinks(ILinkMVCBundle linkBundle,
			Map<Object, IDiagramAnchorageMVCBundle> diagramAnchorageBundles,
			Map<Object, IContentAnchorageMVCBundle> contentAnchorageBundles) {
		Object link = linkBundle.getModel();
		Object startModel = modelFactory.getLinkStart(link);
		if(startModel == null) {
			throw new IllegalArgumentException("Link start model element cannot be null!");
		}
		IAnchorageMVCBundle startBundle = getBundleForAnchorageModel(startModel,
				diagramAnchorageBundles, contentAnchorageBundles);
		Object endModel = modelFactory.getLinkEnd(link);
		if(endModel == null) {
			throw new IllegalArgumentException("Link target model element cannot be null!");
		}
		IAnchorageMVCBundle endBundle = getBundleForAnchorageModel(endModel,
				diagramAnchorageBundles, contentAnchorageBundles);
		linkBundle.updateLinkBundle(startBundle, endBundle);
	}

	/** Returns the {@link MVCBundleBase} of the given anchorage node's model. */
	private IAnchorageMVCBundle getBundleForAnchorageModel(Object m,
			Map<Object, IDiagramAnchorageMVCBundle> diagramAnchorageBundles,
			Map<Object, IContentAnchorageMVCBundle> contentAnchorageBundles) {
		IAnchorageMVCBundle result = diagramAnchorageBundles.get(m);
		if(result != null) {
			return result;
		}
		result = contentAnchorageBundles.get(m);
		if(result != null) {
			return result;
		}
		throw new IllegalArgumentException(
				"Anchorage node model is neither a diagram nor a content anchorage.");
	}

	/** Updates every visual (and only its appearance, not the structure). */
	/* package */ void updateAllVisuals() {
		layers.setScale(viewer.getFeatures().getCurrentZoomFactor());
		viewer.updateGridCanvasSize();
		gridCanvasVisual.updateNodes(layers);
		for(IContentMVCBundle b : contentBundles.values()) {
			if(b instanceof IContentMVCBundleWithParent) {
				IContentMVCBundleWithParent wp = (IContentMVCBundleWithParent)b;
				if(wp.getParentBundle() == null) {
					updateChildContentVisualsRecursively(b);
				}
			} else {
				updateChildContentVisualsRecursively(b);
			}
		}
		for(IDiagramAnchorageMVCBundle b : diagramAnchorageBundles.values()) {
			b.getVisual().updateNodes(layers);
		}
		for(IContentAnchorageMVCBundle b : contentAnchorageBundles.values()) {
			b.getVisual().updateNodes(layers);
		}
		for(ILinkMVCBundle b : linkBundles.values()) {
			b.getVisual().updateNodes(layers);
		}
	}

	/** Updates all child content visuals recursively (but does not update anchorages and links). */
	private void updateChildContentVisualsRecursively(IContentMVCBundle bundle) {
		bundle.getVisual().updateNodes(layers);
		for(IContentMVCBundle cb : bundle.getChildren()) {
			updateChildContentVisualsRecursively(cb);
		}
	}

	/** Updates the nodes selectively adhering to the relation given by the {@link IMVCBundle}s. */
	/* package */ void updateVisual(IMVCBundle bundle) {
		if(bundle instanceof IContentMVCBundle) {
			updateContentVisual((IContentMVCBundle)bundle);
		} else if(bundle instanceof IAnchorageMVCBundle) {
			updateAnchorage((IAnchorageMVCBundle)bundle);
		} else if(bundle instanceof ILinkMVCBundle) {
			bundle.getVisual().updateNodes(layers);
		} else if(bundle instanceof IDiagramMVCBundle) {
			updateAllVisuals();
		}
	}

	/** Updates the nodes selectively adhering to the relation given by the {@link IMVCBundle}s. */
	/* package */ void updateModelVisual(Object model) {
		if(contentBundles.containsKey(model)) {
			updateContentVisual(contentBundles.get(model));
		} else if(contentAnchorageBundles.containsKey(model)) {
			updateAnchorage(contentAnchorageBundles.get(model));
		} else if(diagramAnchorageBundles.containsKey(model)) {
			updateAnchorage(diagramAnchorageBundles.get(model));
		} else if(linkBundles.containsKey(model)) {
			linkBundles.get(model).getVisual().updateNodes(layers);
		} else if(diagramBundle.getModel() == model) {
			updateAllVisuals();
		}
	}

	/** Updates the given {@link IContentMVCBundle}. */
	private void updateContentVisual(IContentMVCBundle bundle) {
		bundle.getVisual().updateNodes(layers);
		for(IContentAnchorageMVCBundle cab : bundle.getAnchorages()) {
			updateAnchorage(cab);
		}
		for(IContentMVCBundle cb : bundle.getChildren()) {
			updateContentVisual(cb);
		}
	}

	/** Updates the given {@link IContentAnchorageMVCBundle}. */
	private void updateAnchorage(IAnchorageMVCBundle bundle) {
		bundle.getVisual().updateNodes(layers);
		for(ILinkMVCBundle link : bundle.getIncomingLinks()) {
			link.getVisual().updateNodes(layers);
		}
		for(ILinkMVCBundle link : bundle.getOutgoingLinks()) {
			link.getVisual().updateNodes(layers);
		}
	}

	/** Returns the root model element of this diagram. */
	/* package */ Object getRootModel() {
		return modelFactory.getRootModel();
	}

	/** Returns the current selection. */
	public DiagramViewerSelection getSelection() {
		return new DiagramViewerSelection(primarySelectedBundle, secondarySelectedBundles);
	}

	/** Returns whether the given bundle is selected as primary. */
	public boolean isPrimarySelected(IMVCBundle bundle) {
		return primarySelectedBundle == bundle;
	}

	/** Returns whether the given bundle is selected as secondary. */
	public boolean isSecondarySelected(IMVCBundle bundle) {
		return secondarySelectedBundles.contains(bundle);
	}

	/** Performs shift selection of the given bundle. */
	public void handleCtrlSelectionOf(IMVCBundle sel) {
		if(primarySelectedBundle == sel || secondarySelectedBundles.contains(sel)) {
			removeSelectedMVCBundle(sel);
		} else {
			addSelectedMVCBundle(sel);
		}
	}

	/** Selects all content, diagram anchorage and link {@link IMVCBundle}s. */
	public void selectAll() {
		secondarySelectedBundles.addAll(contentBundles.values());
		secondarySelectedBundles.addAll(diagramAnchorageBundles.values());
		secondarySelectedBundles.addAll(linkBundles.values());
		if(secondarySelectedBundles.isEmpty()) {
			if(primarySelectedBundle != null) {
				primarySelectedBundle.removeTag(PRIMARY_SELECTION_TAG);
			}
			primarySelectedBundle = null;
		} else {
			primarySelectedBundle = secondarySelectedBundles.remove(0);
			primarySelectedBundle.removeTag(SECONDARY_SELECTION_TAG);
			primarySelectedBundle.addTag(PRIMARY_SELECTION_TAG);
		}
		for(IMVCBundle s : secondarySelectedBundles) {
			s.removeTag(PRIMARY_SELECTION_TAG);
			s.addTag(SECONDARY_SELECTION_TAG);
		}
		updateAllVisuals();
		fireSelectionChanged();
	}
}
