/*******************************************************************************
 * Copyright (c) 2019, 2020 fortiss GmbH. 
 * 
 * This program and the accompanying materials are made available under the
 * terms of the Apache License, Version 2.0 which is available at
 * https://www.apache.org/licenses/LICENSE-2.0.
 *
 * SPDX-License-Identifier: Apache-2.0
 *******************************************************************************/
package org.fortiss.tooling.common.ui.javafx.lwfxef;

import java.util.List;

import javafx.geometry.Bounds;
import javafx.scene.Group;
import javafx.scene.Node;
import javafx.scene.layout.Pane;
import javafx.scene.layout.Region;
import javafx.scene.paint.Color;
import javafx.scene.paint.Paint;
import javafx.scene.shape.Circle;
import javafx.scene.shape.ClosePath;
import javafx.scene.shape.CubicCurve;
import javafx.scene.shape.CubicCurveTo;
import javafx.scene.shape.Ellipse;
import javafx.scene.shape.Line;
import javafx.scene.shape.LineTo;
import javafx.scene.shape.MoveTo;
import javafx.scene.shape.Path;
import javafx.scene.shape.PathElement;
import javafx.scene.shape.QuadCurve;
import javafx.scene.shape.QuadCurveTo;
import javafx.scene.shape.Rectangle;
import javafx.scene.shape.Shape;
import javafx.scene.text.Text;

/** Export the {@link DiagramViewer} to a SVG file. */
final class SVGExporter {
	/** The diagram viewer node to be exported. */
	private final Pane diagramNode;
	/** The width of the diagram computed from the nodes. */
	private double width = 0;
	/** The height of the diagram computed from the nodes. */
	private double height = 0;

	/** Constructor. */
	public SVGExporter(DiagramViewer viewer) {
		// first element of the diagram root scrolled pane is the content pane, see DiagramViewer
		this.diagramNode = (Pane)viewer.getVisualNode().getChildren().get(0);
	}

	/** Exports the diagram to the target file. */
	public String export() {
		width = height = 0;
		// first convert the nodes in order to compute width and height
		StringBuilder content = new StringBuilder();
		for(Node n : diagramNode.getChildrenUnmodifiable()) {
			transform(n, content);
		}

		// then build an SVG XML file
		StringBuilder sb = new StringBuilder();
		sb.append("<?xml version=\"1.0\" standalone=\"no\"?>\n");

		sb.append("<!DOCTYPE svg PUBLIC \"-//W3C//DTD SVG 1.1//EN\" ");
		sb.append("\"http://www.w3.org/Graphics/SVG/1.1/DTD/svg11.dtd\">\n");

		sb.append("<svg ");
		setProp("width", width, sb);
		setProp("height", height, sb);
		setProp("version", "1.1", sb);
		setProp("xmlns", "http://www.w3.org/2000/svg", sb);
		setProp("xmlns:xlink", "http://www.w3.org/1999/xlink", sb);
		sb.append(">\n");

		sb.append(content);
		sb.append("</svg>\n");
		return sb.toString();
	}

	/** Transforms from {@link Node} to a SVG element. */
	private void transform(Node node, StringBuilder sb) {
		Bounds b = node.getBoundsInParent();
		if(b.getMaxX() > width) {
			width = b.getMaxX();
		}
		if(b.getMaxY() > height) {
			height = b.getMaxY();
		}
		if(node instanceof Region) {
			transform((Region)node, sb);
		} else if(node instanceof Group) {
			transform((Group)node, sb);
		} else if(node instanceof Rectangle) {
			transform((Rectangle)node, sb);
		} else if(node instanceof Ellipse) {
			transform((Ellipse)node, sb);
		} else if(node instanceof Circle) {
			transform((Circle)node, sb);
		} else if(node instanceof QuadCurve) {
			transform((QuadCurve)node, sb);
		} else if(node instanceof CubicCurve) {
			transform((CubicCurve)node, sb);
		} else if(node instanceof Line) {
			transform((Line)node, sb);
		} else if(node instanceof Text) {
			transform((Text)node, sb);
		} else if(node instanceof Path) {
			transform((Path)node, sb);
		}
		// silently ignore the rest
	}

	/** Writes a SVG path element. */
	private static void writePath(StringBuilder sb, String op, double... coords) {
		sb.append(op).append(' ');
		for(double d : coords) {
			sb.append(d).append(' ');
		}
	}

	/** Transforms from {@link QuadCurve} to SVG Path element. */
	private void transform(CubicCurve curve, StringBuilder sb) {
		StyleHelper st = new StyleHelper(curve);
		if(st.isTransparent()) {
			return;
		}
		StringBuilder db = new StringBuilder();
		db.append("M ").append(curve.getStartX()).append(' ').append(curve.getStartY()).append(' ');
		db.append("C ").append(curve.getControlX1()).append(' ').append(curve.getControlY1())
				.append(' ');
		db.append(curve.getControlX2()).append(' ').append(curve.getControlY2()).append(' ');
		db.append(curve.getEndX()).append(' ').append(curve.getEndY());
		sb.append("<path ");
		setProp("d", db.toString(), sb);
		st.applySVGStyle(sb);
		sb.append("/>\n");
	}

	/** Transforms from {@link QuadCurve} to SVG Path element. */
	private void transform(QuadCurve curve, StringBuilder sb) {
		StyleHelper st = new StyleHelper(curve);
		if(st.isTransparent()) {
			return;
		}
		StringBuilder db = new StringBuilder();
		db.append("M ").append(curve.getStartX()).append(' ').append(curve.getStartY()).append(' ');
		db.append("Q ").append(curve.getControlX()).append(' ').append(curve.getControlY())
				.append(' ');
		db.append(curve.getEndX()).append(' ').append(curve.getEndY());
		sb.append("<path ");
		setProp("d", db.toString(), sb);
		st.applySVGStyle(sb);
		sb.append("/>\n");
	}

	/** Converts {@link PathElement}s to SVG D specification. */
	private static String convertPathToD(Path path) {
		StringBuilder db = new StringBuilder();
		for(PathElement pe : path.getElements()) {
			if(pe instanceof MoveTo) {
				MoveTo mt = (MoveTo)pe;
				writePath(db, "M", mt.getX(), mt.getY());
			} else if(pe instanceof LineTo) {
				LineTo lt = (LineTo)pe;
				writePath(db, "L", lt.getX(), lt.getY());
			} else if(pe instanceof QuadCurveTo) {
				QuadCurveTo qc = (QuadCurveTo)pe;
				writePath(db, "Q", qc.getControlX(), qc.getControlY(), qc.getX(), qc.getY());
			} else if(pe instanceof CubicCurveTo) {
				CubicCurveTo cc = (CubicCurveTo)pe;
				writePath(db, "C", cc.getControlX1(), cc.getControlY1(), cc.getControlX2(),
						cc.getControlY2(), cc.getX(), cc.getY());
			} else if(pe instanceof ClosePath) {
				writePath(db, "Z");
			}
		}
		return db.toString();
	}

	/** Transforms from {@link Path} to SVG Path element. */
	private void transform(Path path, StringBuilder sb) {
		StyleHelper st = new StyleHelper(path);
		if(st.isTransparent()) {
			return;
		}
		sb.append("<path ");
		setProp("d", convertPathToD(path), sb);
		st.applySVGStyle(sb);
		sb.append("/>\n");
	}

	/** Transforms from {@link Text} to SVG Path element. */
	private void transform(Text text, StringBuilder sb) {
		StyleHelper st = new StyleHelper(text);
		if(st.isTransparent()) {
			return;
		}
		sb.append("<text ");
		double yBaselineOffsetFromTop =
				text.getBoundsInParent().getHeight() - text.getBaselineOffset();
		setProp("x", text.getX(), sb);
		// SVG uses the y-coordinate of the baseline
		setProp("y", text.getY() + yBaselineOffsetFromTop, sb);
		setProp("font-size", text.getFont().getSize(), sb);
		setProp("font-family", text.getFont().getName() + ", Arial, Helvetica, sans-serif", sb);
		st.applySVGStyle(sb);
		sb.append(">\n");
		sb.append(text.getText());
		sb.append("</text>\n");
	}

	/** Transforms from {@link Line} to SVG element. */
	private void transform(Line line, StringBuilder sb) {
		StyleHelper st = new StyleHelper(line);
		if(st.isTransparent()) {
			return;
		}
		sb.append("<line ");
		setProp("x1", line.getStartX(), sb);
		setProp("y1", line.getStartY(), sb);
		setProp("x2", line.getEndX(), sb);
		setProp("y2", line.getEndY(), sb);
		st.applySVGStyle(sb);
		sb.append("/>\n");
	}

	/** Transforms from {@link Circle} to SVG element. */
	private void transform(Circle circ, StringBuilder sb) {
		StyleHelper st = new StyleHelper(circ);
		if(st.isTransparent()) {
			return;
		}
		sb.append("<circle ");
		setProp("cx", circ.getCenterX(), sb);
		setProp("cy", circ.getCenterY(), sb);
		setProp("r", circ.getRadius(), sb);
		st.applySVGStyle(sb);
		sb.append("/>\n");
	}

	/** Transforms from {@link Ellipse} to SVG element. */
	private void transform(Ellipse ell, StringBuilder sb) {
		StyleHelper st = new StyleHelper(ell);
		if(st.isTransparent()) {
			return;
		}
		sb.append("<ellipse ");
		setProp("cx", ell.getCenterX(), sb);
		setProp("cy", ell.getCenterY(), sb);
		setProp("rx", ell.getRadiusX(), sb);
		setProp("ry", ell.getRadiusY(), sb);
		st.applySVGStyle(sb);
		sb.append("/>\n");
	}

	/** Transforms from {@link Rectangle} to SVG Rect element. */
	private void transform(Rectangle rect, StringBuilder sb) {
		StyleHelper st = new StyleHelper(rect);
		if(st.isTransparent()) {
			return;
		}
		sb.append("<rect ");
		setProp("x", rect.getX(), sb);
		setProp("y", rect.getY(), sb);
		setProp("rx", rect.getArcWidth(), sb);
		setProp("ry", rect.getArcHeight(), sb);
		setProp("width", rect.getWidth(), sb);
		setProp("height", rect.getHeight(), sb);
		st.applySVGStyle(sb);
		sb.append("/>\n");
	}

	/** Transforms from {@link Group} to SVG G element. */
	private void transform(Group g, StringBuilder sb) {
		transformToG(null, g.getChildrenUnmodifiable(), sb);
	}

	/** Write a SVG G element. */
	private void transformToG(String transform, List<Node> children, StringBuilder sb) {
		sb.append("<g ");
		if(transform != null) {
			sb.append("transform=\"").append(transform).append("\" ");
		}
		sb.append(">\n");
		for(Node n : children) {
			if(n.isVisible()) {
				transform(n, sb);
			}
		}
		sb.append("</g>\n");
	}

	/** Transforms from {@link Region} to SVG G element. */
	private void transform(Region r, StringBuilder sb) {
		Bounds b = r.getBoundsInParent();
		String transform = "translate(" + b.getMinX() + "," + b.getMinY() + ")";
		transformToG(transform, r.getChildrenUnmodifiable(), sb);
	}

	/** Writes the given property and value to the string builder. */
	private static void setProp(String prop, String value, StringBuilder sb) {
		sb.append(prop).append("=\"").append(value).append("\" ");
	}

	/** Writes the given property and value to the string builder. */
	private static void setProp(String prop, double value, StringBuilder sb) {
		sb.append(prop).append("=\"").append(value).append("\" ");
	}

	/** Helper class for handling fill and stroke styles. */
	private static class StyleHelper {
		/** RGB hex of the fill color. */
		private String fill;
		/** Opacity of the fill color. */
		private double fillAlpha;
		/** RGB hex of the stroke color. */
		private String stroke;
		/** Opacity of the stroke color. */
		private double strokeAlpha;
		/** Width of the stroke. */
		private double strokeWidth;

		/** Constructor. */
		public StyleHelper(Paint back, Paint fore, double sw) {
			fill = rgbHex(back);
			fillAlpha = alpha(back);
			stroke = rgbHex(fore);
			strokeAlpha = alpha(fore);
			strokeWidth = sw;
		}

		/** Constructor. */
		public StyleHelper(Shape shape) {
			this(shape.getFill(), shape.getStroke(), shape.getStrokeWidth());
		}

		/** Returns the SVG string for this style. */
		public void applySVGStyle(StringBuilder sb) {
			setProp("fill", fill != null ? fill : "none", sb);
			setProp("fill-opacity", fillAlpha, sb);
			setProp("stroke", stroke != null ? stroke : "none", sb);
			setProp("stroke-width", strokeWidth, sb);
			setProp("stroke-opacity", strokeAlpha, sb);
		}

		/** Returns whether the style has transparent colors. */
		public boolean isTransparent() {
			return (fill == null && stroke == null) || (fillAlpha == 0 && strokeAlpha == 0);
		}

		/** Converts from {@link Paint} to SVG hex-decimal color specification. */
		private static String rgbHex(Paint p) {
			if(p instanceof Color) {
				Color c = (Color)p;
				int r = (int)(c.getRed() * 255);
				int g = (int)(c.getGreen() * 255);
				int b = (int)(c.getBlue() * 255);
				return String.format("#%02X%02X%02X", r, g, b);
			}
			return null;
		}

		/** Converts from Paint to alpha value. */
		private static double alpha(Paint p) {
			if(p instanceof Color) {
				return ((Color)p).getOpacity();
			}
			return 0.0;
		}
	}
}
