/*******************************************************************************
 * Copyright (c) 2019, 2020 fortiss GmbH. 
 * 
 * This program and the accompanying materials are made available under the
 * terms of the Apache License, Version 2.0 which is available at
 * https://www.apache.org/licenses/LICENSE-2.0.
 *
 * SPDX-License-Identifier: Apache-2.0
 *******************************************************************************/
package org.fortiss.tooling.common.ui.javafx.lwfxef.controller.base;

import static org.fortiss.tooling.common.ui.javafx.lwfxef.FeedbackChange.locationFeedbackChange;

import org.fortiss.tooling.common.ui.javafx.lwfxef.DiagramCoordinate;
import org.fortiss.tooling.common.ui.javafx.lwfxef.FeedbackChange;
import org.fortiss.tooling.common.ui.javafx.lwfxef.change.Change;
import org.fortiss.tooling.common.ui.javafx.lwfxef.controller.IDragController;
import org.fortiss.tooling.common.ui.javafx.lwfxef.mvc.IMVCBundle;

import javafx.scene.Node;

/** Base implementation of {@link IDragController}. */
public abstract class DragControllerBase implements IDragController {
	/** Stores the start bundle of the drag gesture. */
	protected IMVCBundle startBundle = null;
	/** Stores the start node of the drag gesture. */
	protected Node startNode;
	/** Stores the start location of the drag gesturein diagram coordinates. */
	protected DiagramCoordinate startDiagramLocation = null;

	/** {@inheritDoc} */
	@Override
	public final void dragStarted(IMVCBundle bundle, Node node, DiagramCoordinate diagramLocation) {
		this.startBundle = bundle;
		this.startNode = node;
		this.startDiagramLocation = diagramLocation;
		safeDragStarted(bundle, node, startDiagramLocation);
	}

	/**
	 * Informs the drag controller that a drag has started.
	 * 
	 * @param bundle
	 *            the final {@link IMVCBundle}
	 * @param node
	 *            the final node
	 * @param diagramLocation
	 *            the current absolute mouse location
	 */
	protected void safeDragStarted(IMVCBundle bundle, Node node,
			DiagramCoordinate diagramLocation) {
		// the default does nothing
	}

	/** {@inheritDoc} */
	@Override
	public final void dragInProgress(IMVCBundle bundle, Node node,
			DiagramCoordinate diagramLocation) {
		// filter drag events that arrive before drag started was called.
		if(isDragInProgress()) {
			safeDragInProgress(bundle, node, diagramLocation);
		}
	}

	/**
	 * Updates the drag controller with the new drag location guaranteeing that its only called
	 * after {@link #dragStarted(IMVCBundle, Node, DiagramCoordinate)} was called and before
	 * {@link #dragCompleted(IMVCBundle, Node, DiagramCoordinate)} is
	 * called.
	 * 
	 * @param bundle
	 *            the current {@link IMVCBundle}
	 * @param node
	 *            the current node under the drag gesture pointer
	 * @param diagramLocation
	 *            the current diagram location of the drag gesture
	 */
	protected void safeDragInProgress(IMVCBundle bundle, Node node,
			DiagramCoordinate diagramLocation) {
		// the default does nothing
	}

	/** {@inheritDoc} */
	@Override
	public final Change dragCompleted(IMVCBundle bundle, Node node,
			DiagramCoordinate diagramLocation) {
		Change result = null;
		if(isDragInProgress()) {
			result = safeDragCompleted(bundle, node, diagramLocation);
		}
		startBundle = null;
		startNode = null;
		startDiagramLocation = null;
		return result;
	}

	/**
	 * Updates the drag controller with the final drag location guaranteeing that its only called
	 * after {@link #dragStarted(IMVCBundle, Node, DiagramCoordinate)} was called and any number of
	 * {@link #safeDragInProgress(IMVCBundle, Node, DiagramCoordinate)} occurred.
	 * 
	 * @param bundle
	 *            the final {@link IMVCBundle}
	 * @param node
	 *            the final node under the gesture pointer
	 * @param diagramLocation
	 *            the final diagram location of the drag gesture
	 * @return the model change which was caused by this drag operation.
	 */
	protected Change safeDragCompleted(IMVCBundle bundle, Node node,
			DiagramCoordinate diagramLocation) {
		// the default does nothing
		return null;
	}

	/**
	 * Returns whether a drag gesture is in progress, i.e., dragStarted() was called and
	 * dragCompleted() was not yet called.
	 */
	protected final boolean isDragInProgress() {
		return startBundle != null;
	}

	/** Returns the current drag delta. */
	protected final FeedbackChange getDragDelta(DiagramCoordinate endDiagramLocation) {
		double dx = endDiagramLocation.getX() - startDiagramLocation.getX();
		double dy = endDiagramLocation.getY() - startDiagramLocation.getY();
		return locationFeedbackChange(dx, dy);
	}
}
