/*******************************************************************************
 * Copyright (c) 2019, 2020 fortiss GmbH. 
 * 
 * This program and the accompanying materials are made available under the
 * terms of the Apache License, Version 2.0 which is available at
 * https://www.apache.org/licenses/LICENSE-2.0.
 *
 * SPDX-License-Identifier: Apache-2.0
 *******************************************************************************/
package org.fortiss.tooling.common.ui.javafx.lwfxef.controller.rectangular;

import static java.lang.Math.hypot;

import org.apache.commons.lang3.tuple.ImmutablePair;
import org.apache.commons.lang3.tuple.Pair;
import org.fortiss.tooling.common.ui.javafx.lwfxef.DiagramCoordinate;
import org.fortiss.tooling.common.ui.javafx.lwfxef.controller.base.ContentAnchorageMoveControllerBase;
import org.fortiss.tooling.common.ui.javafx.lwfxef.mvc.IContentAnchorageMVCBundle;
import org.fortiss.tooling.common.ui.javafx.lwfxef.visual.IContentAnchorageVisual;
import org.fortiss.tooling.common.ui.javafx.lwfxef.visual.IVisual;
import org.fortiss.tooling.common.ui.javafx.lwfxef.visual.rectangular.RectangularBorderLocation;
import org.fortiss.tooling.common.ui.javafx.lwfxef.visual.rectangular.RhomboidContentVisualBase;

import javafx.geometry.Dimension2D;
import javafx.geometry.Point2D;
import javafx.geometry.Rectangle2D;
import javafx.geometry.Side;

/**
 * Default implementation for moving content anchorages on the border of a rhomboid content
 * model element visual.
 */
public class RhomboidContentAnchorageMoveController extends ContentAnchorageMoveControllerBase {
	/** The controller of the rhomboid content visual. */
	private final RhomboidResizableContentControllerBase contentController;

	/** Constructor. */
	public RhomboidContentAnchorageMoveController(RhomboidResizableContentControllerBase controller,
			IContentAnchorageMVCBundle anchorage) {
		super(anchorage);
		this.contentController = controller;
		if(!(contentController.getVisual() instanceof RhomboidContentVisualBase)) {
			throw new IllegalArgumentException("Expected RhomboidContentVisualBase.");
		}
	}

	/** Returns the location clamped to the border of the parent visual. */
	private Pair<Side, Double> getLocationOnBorder(IContentAnchorageMVCBundle anchorage,
			Point2D delta) {
		IContentAnchorageVisual anchorageVisual = anchorage.getVisual();
		Rectangle2D anchorageBounds = anchorageVisual.getModelBounds();
		double awidth = anchorageBounds.getWidth();
		double aheight = anchorageBounds.getHeight();

		IVisual contentVisual = contentController.getVisual();
		Rectangle2D contentBounds = contentVisual.getModelBounds();
		double rhomboidOffset = ((RhomboidContentVisualBase)contentVisual).getOffset();

		double topX = contentBounds.getMinX();
		double indicatorX = topX + rhomboidOffset + delta.getX() + awidth / 2;
		double widthLessOffset = contentBounds.getWidth() - rhomboidOffset;
		double topY = contentBounds.getMinY();
		double indicatorY = topY + delta.getY() + aheight / 2;
		double botY = contentBounds.getMaxY();

		Side resultSide = Side.TOP;
		double resultOffset = rhomboidOffset;
		double hypo = Double.MAX_VALUE; // hypot(indicatorX - topX, indicatorY - topY);

		// search TOP and BOTTOM locations
		for(double fx = 0; fx <= 1; fx += 0.25) {
			double tx = topX + rhomboidOffset + fx * widthLessOffset;
			double topHypo = hypot(indicatorX - tx, indicatorY - topY);
			if(topHypo < hypo) {
				hypo = topHypo;
				resultOffset = tx - topX;
				resultSide = Side.TOP;
			}
			double bx = topX + fx * widthLessOffset;
			double botHypo = hypot(indicatorX - bx, indicatorY - botY);
			if(botHypo < hypo) {
				hypo = botHypo;
				resultOffset = bx - topX;
				resultSide = Side.BOTTOM;
			}
		}

		double height2 = contentBounds.getHeight() / 2;
		double halfHeightY = topY + height2;
		// search LEFT center
		double lx = topX + rhomboidOffset / 2;
		double leftHypo = hypot(indicatorX - lx, indicatorY - halfHeightY);
		if(leftHypo < hypo) {
			hypo = leftHypo;
			resultOffset = height2;
			resultSide = Side.LEFT;
		}
		// search RIGHT center
		double rx = topX + contentBounds.getWidth() - rhomboidOffset / 2;
		double rightHypo = hypot(indicatorX - rx, indicatorY - halfHeightY);
		if(rightHypo < hypo) {
			resultOffset = height2;
			resultSide = Side.RIGHT;
		}
		// return closest test point
		return new ImmutablePair<Side, Double>(resultSide, resultOffset - awidth / 2);
	}

	/** Allows sub-classes to override the default span calculation. */
	protected Dimension2D overrideBorderSnap(Dimension2D anchorageBounds) {
		return anchorageBounds;
	}

	/** {@inheritDoc} */
	@Override
	protected Point2D getAnchorageFeedbackLocation(DiagramCoordinate delta) {
		RhomboidContentVisualBase contentVisual =
				(RhomboidContentVisualBase)contentController.getVisual();
		Rectangle2D contentBounds = contentVisual.getModelBounds();
		Rectangle2D anchorageBounds = anchorage.getVisual().getModelBounds();
		Dimension2D anchorageSize =
				new Dimension2D(anchorageBounds.getWidth(), anchorageBounds.getHeight());
		double rhomboidOffset = contentVisual.getOffset();

		Pair<Side, Double> lob = getLocationOnBorder(anchorage, delta);
		Side side = lob.getLeft();
		if(side == Side.LEFT) {
			double lx = rhomboidOffset / 2 - anchorageSize.getWidth() / 2;
			double ly = (contentBounds.getHeight() - anchorageSize.getHeight()) / 2;
			return new DiagramCoordinate(lx, ly);
		} else if(side == Side.RIGHT) {
			double lx = contentBounds.getWidth() - (rhomboidOffset + anchorageSize.getWidth()) / 2;
			double ly = (contentBounds.getHeight() - anchorageSize.getHeight()) / 2;
			return new DiagramCoordinate(lx, ly);
		}
		RectangularBorderLocation rbl =
				new RectangularBorderLocation(side, lob.getRight(), contentBounds, anchorageSize);
		return rbl.getLocation();
	}

	/** {@inheritDoc} */
	@Override
	protected void moveAnchorage(DiagramCoordinate delta) {
		Pair<Side, Double> lob = getLocationOnBorder(anchorage, delta);
		contentController.moveAnchorageToSideOffset(anchorage, lob.getLeft(), lob.getRight());
	}
}
