/*******************************************************************************
 * Copyright (c) 2019, 2020 fortiss GmbH. 
 * 
 * This program and the accompanying materials are made available under the
 * terms of the Apache License, Version 2.0 which is available at
 * https://www.apache.org/licenses/LICENSE-2.0.
 *
 * SPDX-License-Identifier: Apache-2.0
 *******************************************************************************/
package org.fortiss.tooling.common.ui.javafx.lwfxef.visual.rectangular;

import static org.fortiss.tooling.common.ui.javafx.lwfxef.visual.rectangular.RectangularBorderLocation.getClosestLocationOnBounds;

import org.fortiss.tooling.common.ui.javafx.lwfxef.DiagramCoordinate;
import org.fortiss.tooling.common.ui.javafx.lwfxef.DiagramLayers;
import org.fortiss.tooling.common.ui.javafx.lwfxef.EDragGesture;
import org.fortiss.tooling.common.ui.javafx.lwfxef.model.layout.ISideLayout;
import org.fortiss.tooling.common.ui.javafx.lwfxef.mvc.IContentMVCBundle;
import org.fortiss.tooling.common.ui.javafx.lwfxef.visual.IContentAnchorageVisual;
import org.fortiss.tooling.common.ui.javafx.lwfxef.visual.base.ContentVisualBase;

import javafx.collections.ObservableList;
import javafx.geometry.Bounds;
import javafx.geometry.Dimension2D;
import javafx.geometry.Rectangle2D;
import javafx.geometry.Side;
import javafx.scene.Node;
import javafx.scene.paint.Color;
import javafx.scene.shape.ClosePath;
import javafx.scene.shape.LineTo;
import javafx.scene.shape.MoveTo;
import javafx.scene.shape.Path;
import javafx.scene.shape.PathElement;

/** Base class for {@link ContentVisualBase content visuals} depicted by a diamond shape. */
public abstract class DiamondContentVisualBase extends ContentVisualBase {
	/** The default text anchor location. */
	private static final double DEFAULT_TEXTANCHOR_DIVIDER = 5;

	/** Constructor. */
	public DiamondContentVisualBase(IContentMVCBundle mvcb) {
		super(mvcb, new Path(), new Path());
	}

	/** {@inheritDoc} */
	@Override
	public void updateNodes(DiagramLayers layers) {
		super.updateNodes(layers);

		Rectangle2D bounds = getCurrentBounds();
		double leftX = bounds.getMinX();
		double upperY = bounds.getMinY();
		double width = bounds.getWidth();
		double height = bounds.getHeight();
		if(enableVisual()) {
			Path path = getVisualPathShape();
			path.setOpacity(getOpacity());
			path.setStroke(getBorderColor());
			path.setStrokeWidth(getBorderWidth());
			path.setStrokeType(getBorderType());
			path.setFill(getFillColor());
			makePath(leftX, upperY, width, height, path);
		}

		if(enableInteraction()) {
			Path ha = getHitAreaPathShape();
			double lnk = getHitAreaStartLinkSize();
			double resize = getHitAreaResizeSize();
			double lr = lnk + resize;
			makePath(leftX - lr, upperY - lr, width + 2 * lr, height + 2 * lr, ha);
			Color interactionColor = getInteractionColor();
			ha.setStroke(interactionColor);
			ha.setFill(interactionColor);
		}
	}

	/** Creates the rhomboid shape. */
	private void makePath(double x, double y, double w, double h, Path p) {
		double w2 = w / 2;
		double h2 = h / 2;
		ObservableList<PathElement> pes = p.getElements();
		pes.clear();
		pes.add(new MoveTo(x, y + h2));
		pes.add(new LineTo(x + w2, y + h));
		pes.add(new LineTo(x + w, y + h2));
		pes.add(new LineTo(x + w2, y));
		pes.add(new ClosePath());
	}

	/** {@inheritDoc} */
	@Override
	public DiagramCoordinate getAnchorageLocation(IContentAnchorageVisual visual) {
		Rectangle2D pb = getCurrentBounds();
		double pbw2 = pb.getWidth() / 2;
		double pbh2 = pb.getHeight() / 2;
		Dimension2D anchorageSize = visual.getDimensions();
		double aw2 = anchorageSize.getWidth() / 2;
		double ah2 = anchorageSize.getHeight() / 2;
		Side side = visual.getLayout(ISideLayout.class).getSide();
		// Side.TOP
		double x = pbw2 - aw2;
		double y = -ah2;
		if(side == Side.BOTTOM) {
			y = pb.getHeight() - ah2;
		} else if(side == Side.LEFT) {
			x = -aw2;
			y = pbh2 - ah2;
		} else if(side == Side.RIGHT) {
			x = pb.getWidth() - aw2;
			y = pbh2 - ah2;
		}
		return new DiagramCoordinate(x, y);
	}

	/** {@inheritDoc} */
	@Override
	public DiagramCoordinate getLinkAnchorage(DiagramCoordinate indication) {
		Rectangle2D b = getCurrentBounds();
		Dimension2D dummy = new Dimension2D(1, 1);
		RectangularBorderLocation loc = getClosestLocationOnBounds(
				indication.add(-b.getMinX(), -b.getMinY()), b, dummy, dummy);
		return loc.getLocation().add(b.getMinX(), b.getMinY());
	}

	/** Returns the dimensions of the corner arcs. */
	protected Dimension2D getCornerArcDimensions() {
		return new Dimension2D(15, 15);
	}

	/** {@inheritDoc} */
	@Override
	protected EDragGesture dragGestureHitTest(Node node, DiagramCoordinate diagramLocation) {
		if(node == getHitAreaShape()) {
			double x = diagramLocation.getLocalX(node);
			double y = diagramLocation.getLocalY(node);
			double l = getHitAreaStartLinkSize();
			double r = getHitAreaResizeSize();
			Bounds bounds = node.getBoundsInLocal();
			// check for move area
			double inset = l + r;
			Rectangle2D moveArea = new Rectangle2D(inset, inset, bounds.getWidth() - 2 * inset,
					bounds.getHeight() - 2 * inset);
			if(moveArea.contains(x, y)) {
				return EDragGesture.MOVE;
			}
			if(x < l || y < l || x > bounds.getWidth() - l || y > bounds.getHeight() - l) {
				return EDragGesture.NEW_LINK;
			}
			if(x > bounds.getWidth() - inset && y >= l && y <= bounds.getHeight() - inset) {
				return EDragGesture.RESIZE_H;
			}
			if(y > bounds.getHeight() - inset && x >= l && x <= bounds.getWidth() - inset) {
				return EDragGesture.RESIZE_V;
			}
			if(x > bounds.getWidth() - inset && y > bounds.getHeight() - inset) {
				return EDragGesture.RESIZE_VH;
			}
			return EDragGesture.NONE;
		}
		if(node == getVisualShape() || node == text) {
			return EDragGesture.MOVE;
		}
		return EDragGesture.NONE;
	}

	/** Returns the visual path shape. */
	protected final Path getVisualPathShape() {
		// wild cast works: see constructor
		return (Path)getVisualShape();
	}

	/** Returns the hit area path shape. */
	protected final Path getHitAreaPathShape() {
		// wild cast works: see constructor
		return (Path)getHitAreaShape();
	}

	/**
	 * Returns the extension of the hit area in pixel used to trigger link creation. This number
	 * specifies how many pixels beyond the visible border are used to trigger link creation by a
	 * mouse drag gesture.
	 */
	protected double getHitAreaStartLinkSize() {
		return getViewer().getFeatures().getMaximumSpacing();
	}

	/** Returns the size of the hit area in pixel used to resize the content visual. */
	protected double getHitAreaResizeSize() {
		return getViewer().getFeatures().getMaximumSpacing();
	}

	/** {@inheritDoc} */
	@Override
	protected DiagramCoordinate getTextAnchorLocation() {
		DiagramCoordinate textAnchorLocation = super.getTextAnchorLocation();
		return textAnchorLocation.add(getCurrentBounds().getWidth() / DEFAULT_TEXTANCHOR_DIVIDER,
				0);
	}
}
