/*-------------------------------------------------------------------------+
| Copyright 2016 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.kernel.ui.internal.introspection.details.handler;

import java.util.List;

import org.conqat.ide.commons.ui.jface.TreeContentProviderBase;
import org.conqat.lib.commons.collections.Pair;
import org.eclipse.jface.action.MenuManager;
import org.eclipse.jface.viewers.ITableLabelProvider;
import org.eclipse.jface.viewers.TableLabelProviderBase;
import org.eclipse.swt.SWT;
import org.eclipse.swt.widgets.Tree;
import org.eclipse.swt.widgets.TreeColumn;
import org.eclipse.ui.dialogs.PatternFilter;
import org.fortiss.tooling.kernel.introspection.items.EObjectAwareIntrospectionDetailsItemBase;

/**
 * Base class for details UI implementations with filtered tree viewer.
 * 
 * @author hoelzl
 */
public abstract class EObjectAwareIntrospectionDetailsUIHandlerBase
		extends IntrospectionDetailsUIHandlerBase {

	/** Creates the default tree columns. */
	protected final void createDefaultTreeColumns(Tree tree, String text0, int width0, String text1,
			int width1, String text2, int width2) {
		TreeColumn col0 = new TreeColumn(tree, SWT.LEFT);
		col0.setText(text0);
		col0.setWidth(width0);

		TreeColumn col1 = new TreeColumn(tree, SWT.LEFT);
		col1.setText(text1);
		col1.setWidth(width1);

		TreeColumn col2 = new TreeColumn(tree, SWT.LEFT);
		col2.setText(text2);
		col2.setWidth(width2);
	}

	/** {@inheritDoc} */
	@Override
	protected ITableLabelProvider createLabelProvider() {
		return new EObjectAwareTableLabelProvider();
	}

	/** {@inheritDoc} */
	@Override
	protected PatternFilter createPatternFilter() {
		PatternFilter pf = new TableViewerPatternFilter();
		pf.setIncludeLeadingWildcard(true);
		return pf;
	}

	/** Table label provider for handler registrations with two classes. */
	protected static class EObjectAwareTableLabelProvider extends TableLabelProviderBase {
		/** {@inheritDoc} */
		@Override
		public String getColumnText(Object element, int columnIndex) {
			if(element instanceof Class) {
				return columnIndex == 0 ? ((Class<?>)element).getSimpleName() : "";
			}
			if(element instanceof Pair) {
				Pair<?, ?> pair = (Pair<?, ?>)element;
				if(pair.getFirst() != null) {
					Object provider = pair.getFirst();
					Class<?> regClass = (Class<?>)pair.getSecond();
					switch(columnIndex) {
						case 0:
							return provider.getClass().getSimpleName();
						case 1:
							return provider.getClass().getName();
						case 2:
							return regClass.getName();
					}
				}
			}
			return "";
		}
	}

	/** Tree content provider for handler registrations with two classes. */
	protected static abstract class EObjectAwareTreeContentProviderBase
			extends TreeContentProviderBase {

		/** Returns the input object for this content provider. */
		protected abstract EObjectAwareIntrospectionDetailsItemBase<?> getInputObject();

		/** {@inheritDoc} */
		@Override
		public Object[] getChildren(Object parentElement) {
			if(parentElement == getInputObject()) {
				return getInputObject().getHandlerKeyClasses().toArray();
			}
			if(parentElement instanceof Class<?>) {
				Class<?> regClass = (Class<?>)parentElement;
				List<?> handlerList = getInputObject().getHandlerList(regClass);
				if(handlerList == null || handlerList.isEmpty()) {
					return new Object[0];
				}
				Pair<?, ?>[] pairs = new Pair<?, ?>[handlerList.size()];
				int idx = 0;
				for(Object handler : handlerList) {
					pairs[idx] = new Pair<Object, Class<?>>(handler, regClass);
					idx++;
				}
				return pairs;
			}
			return new Object[0];
		}
	}

	/** {@inheritDoc} */
	@Override
	protected boolean testSelection(Object selection) {
		return selection instanceof Pair<?, ?>;
	}

	/** {@inheritDoc} */
	@Override
	protected void populateContextMenu(Object selection, MenuManager mgr) {
		Pair<?, ?> pair = (Pair<?, ?>)selection;
		mgr.add(createCopyClassNameAction(pair.getFirst().getClass()));
		mgr.add(createCopyClassNameAction((Class<?>)pair.getSecond()));
	}
}
