/*-------------------------------------------------------------------------+
| Copyright 2015 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.kernel.constraint;

import org.eclipse.emf.ecore.EObject;
import org.fortiss.tooling.kernel.model.IIdLabeled;
import org.fortiss.tooling.kernel.model.INamedElement;

/**
 * Class for implementation of constraint violation message.
 * 
 * @author barner
 */
public class ConstraintMessage {

	/**
	 * Formats the name of {@code element} for the {@link IdUniquenessConstraintViolation} message.
	 */
	private static String formatElementName(EObject element) {
		return formatElementName(element, true);
	}

	/** Helper method to implement {@link #formatElementName(EObject)}. */
	private static String formatElementName(EObject element, boolean offspring) {
		if(element instanceof INamedElement) {
			String elementName = ((INamedElement)element).getName();
			if(elementName != null && !elementName.isEmpty()) {
				return "\"" + elementName + "\"";
			}
		}

		if(element != null && element.eContainer() != null) {
			String elementName = formatElementName(element.eContainer(), false);

			return offspring ? "<unnamed offspring of> " + elementName : elementName;
		}

		return "<unnamed>";
	}

	/** Creates the violation for an {@link IIdLabeled} element that contains a duplicated ID. */
	public static IdUniquenessConstraintViolation<IIdLabeled>
			createDuplicateIdViolation(IIdLabeled element, IIdLabeled duplicate) {
		return new IdUniquenessConstraintViolation<IIdLabeled>(element,
				"The ID " + element.getId() + " of model element " + formatElementName(element) +
						" has already been assigned to model element " +
						formatElementName(duplicate) + ".");
	}

	/**
	 * Creates the violation for an {@link IIdLabeled} whose ID attribute is inconsistent with the
	 * elements persisted id (e.g., {@code xmi:id}).
	 */
	public static IdConsistencyConstraintViolation<IIdLabeled>
			createConsistencyIdViolation(IIdLabeled element, String persistedId) {
		return new IdConsistencyConstraintViolation<IIdLabeled>(element,
				"The ID " + element.getId() + " of model element " + formatElementName(element) +
						" does not match the element's persisted ID " + persistedId + ".");
	}
}
