/*-------------------------------------------------------------------------+
| Copyright 2020 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.kernel.internal;

import static org.fortiss.tooling.kernel.utils.LoggingUtils.error;

import java.util.HashMap;
import java.util.Map;

import org.eclipse.core.runtime.Platform;
import org.fortiss.tooling.kernel.ToolingKernelActivator;
import org.fortiss.tooling.kernel.extension.ICommandLineSwitchHandler;
import org.fortiss.tooling.kernel.service.ICommandLineInterfaceService;

/**
 * Implementation of the {@link ICommandLineInterfaceService}.
 * 
 * @author diewald
 */
public class CommandLineInterfaceService implements ICommandLineInterfaceService {

	/** Singleton instance of the {@link ICommandLineInterfaceService} service. */
	private static CommandLineInterfaceService INSTANCE = new CommandLineInterfaceService();

	/** Guard for executing the CLI switch only once during the applications lifetime. */
	private boolean handlersExecuted = false;

	/** Associates CLI switches with their handlers. */
	private Map<String, ICommandLineSwitchHandler> switchHandlerMap = new HashMap<>();

	/** Hiding constructor. */
	private CommandLineInterfaceService() {
		// Prevent external instantiation.
	}

	/** Returns the {@link CommandLineInterfaceService} singleton. */
	public static synchronized CommandLineInterfaceService getInstance() {
		return INSTANCE;
	}

	/** {@inheritDoc} */
	@Override
	public void executeHandlers() {
		if(handlersExecuted) {
			error(ToolingKernelActivator.getDefault(), "The " +
					CommandLineInterfaceService.class.getSimpleName() +
					" was asked more than one to process command line switches. The subsequent " +
					"calls were NOT executed. Please report this issue to the developers.");
			return;
		}
		handlersExecuted = true;

		String[] args = Platform.getApplicationArgs();
		for(int cliIdx = 0; cliIdx < args.length; ++cliIdx) {
			String cliSwitch = args[cliIdx];
			ICommandLineSwitchHandler cliHandler = switchHandlerMap.get(cliSwitch);
			if(cliHandler != null) {
				String argument =
						cliHandler.hasAdditionalArgument() ? argument = args[++cliIdx] : null;
				cliHandler.handleCLISwitch(argument);
			} else {
				error(ToolingKernelActivator.getDefault(),
						"Could not parse the argument " + cliSwitch + ".");
			}
		}
	}

	/** {@inheritDoc} */
	@Override
	public void registerHandler(String cliSwitch, ICommandLineSwitchHandler cliHandler) {
		if(switchHandlerMap.containsKey(cliSwitch)) {
			ICommandLineSwitchHandler activeHandler = switchHandlerMap.get(cliSwitch);
			error(ToolingKernelActivator.getDefault(),
					"The CLI switch " + cliSwitch + " is already registered with the handler " +
							activeHandler.getClass().getSimpleName() + " and won't be replaced.");
			return;
		}

		switchHandlerMap.put(cliSwitch, cliHandler);
	}
}
