/*-------------------------------------------------------------------------+
| Copyright 2019 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.kernel.utils;

import java.util.ArrayList;
import java.util.EventObject;
import java.util.List;

import org.eclipse.emf.common.command.CommandStackListener;
import org.eclipse.emf.common.notify.Adapter;
import org.eclipse.emf.common.notify.Notification;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.util.EContentAdapter;
import org.fortiss.tooling.kernel.extension.data.ITopLevelElement;
import org.fortiss.tooling.kernel.service.ICommandStackService;
import org.fortiss.tooling.kernel.service.IPersistencyService;

/**
 * Base class for listeners using EMF {@link Adapter}s and {@link ICommandStackService}.
 * <P>
 * Sub-classes should override {@link #updateComplete(List)} in order to be notified when an
 * update cycle has completed (one or more EMF notifications followed by a command stack
 * change).
 * <P>
 * Sub-classes may also provide their custom EMF {@link Adapter}, but must call
 * {@link #addNotification(Notification)} whenever a relevant EMF notification occurs.
 * 
 * @author hoelzl
 */
public abstract class KernelServiceBasedModelListenerBase<T extends EObject>
		implements CommandStackListener {
	/** The persistency service reference. */
	private static final IPersistencyService persistencyService = IPersistencyService.getInstance();
	/** The command stack service. */
	private static final ICommandStackService commandStackService =
			ICommandStackService.getInstance();
	/** List of relevant {@link Notification}s recorded during a change transaction. */
	private final List<Notification> notifications = new ArrayList<>();

	/** The model element. */
	private final T modelElement;
	/** The EMF notification listener. */
	private Adapter emfAdapter = new EContentAdapter() {
		/** {@inheritDoc} */
		@Override
		public void notifyChanged(Notification notification) {
			super.notifyChanged(notification);
			addNotification(notification);
		}
	};

	/** Constructor. */
	public KernelServiceBasedModelListenerBase(T modelElement, Adapter emfAdapter) {
		this.modelElement = modelElement;
		this.emfAdapter = emfAdapter;
	}

	/** Constructor. */
	public KernelServiceBasedModelListenerBase(T modelElement) {
		this.modelElement = modelElement;
	}

	/**
	 * Called when the update cycle is completed.
	 * 
	 * @param notifications
	 *            the list of notifications of the completed cycle
	 */
	public void updateComplete(List<Notification> notifications) {
		// default does nothing.
	}

	/** Adds the given {@link Notification} to list of notifications. */
	protected final void addNotification(Notification event) {
		notifications.add(event);
	}

	/** Registers the listener with the model and the command stack service. */
	public final void register() {
		notifications.clear();
		ITopLevelElement root = persistencyService.getTopLevelElementFor(modelElement);
		commandStackService.addCommandStackListener(root, this);
		modelElement.eAdapters().add(emfAdapter);
	}

	/** Unregisters the listener from the model and the command stack service. */
	public final void unregister() {
		modelElement.eAdapters().remove(emfAdapter);
		ITopLevelElement root = persistencyService.getTopLevelElementFor(modelElement);
		commandStackService.removeCommandStackListener(root, this);
		notifications.clear();
	}

	/** {@inheritDoc} */
	@Override
	public void commandStackChanged(EventObject event) {
		if(notifications.isEmpty()) {
			return;
		}
		updateComplete(notifications);
		notifications.clear();
	}
}
