/*******************************************************************************
 * Copyright (c) 2017, 2018 fortiss GmbH. 
 * 
 * This program and the accompanying materials are made available under the
 * terms of the Apache License, Version 2.0, which is available at 
 * https://www.apache.org/licenses/LICENSE-2.0.
 *
 * SPDX-License-Identifier: Apache-2.0
 *******************************************************************************/
package org.fortiss.tooling.spiderchart.control;

import org.fortiss.tooling.spiderchart.model.SpiderChart;
import org.fortiss.tooling.spiderchart.style.ChartStyle;

import javafx.scene.canvas.Canvas;
import javafx.scene.canvas.GraphicsContext;

/**
 * Represents a {@link Canvas} on which chart will be drawn.
 * 
 * @author hoelzl
 */
final class SpiderChartCanvas {
	/** The canvas. */
	private final Canvas canvas;
	/** The chart style. */
	private final ChartStyle style;
	/** The chart widget. */
	private final SpiderChartControl chartControl;
	/** The title widget. */
	private final SpiderChartTitleControl titleControl;
	/** The legend widget. */
	private final SpiderChartLegendControl legendControl;

	/** Constructor */
	public SpiderChartCanvas(SpiderChart chart, ChartStyle style) {
		this.style = style;
		this.chartControl = new SpiderChartControl(chart, style);
		this.titleControl = new SpiderChartTitleControl(chart, style);
		this.legendControl = new SpiderChartLegendControl(chart, style);

		this.canvas = new Canvas();
		canvas.widthProperty().addListener(evt -> {
			layoutAndPaint();
		});
		canvas.heightProperty().addListener(evt -> {
			layoutAndPaint();
		});
	}

	/** Returns canvas. */
	public Canvas getCanvasControl() {
		return canvas;
	}

	/** Paints the spider chart. */
	private void layoutAndPaint() {
		double width = canvas.getWidth();
		double height = canvas.getHeight();
		layoutChart(width, height);

		GraphicsContext gc = canvas.getGraphicsContext2D();
		gc.clearRect(0, 0, width, height);
		drawChart(gc);
	}

	/** Computes the bounding boxes of the chart components. */
	private void layoutChart(double width, double height) {
		double chartMarginPercent = style.getRelativeMargin();
		double widthMargin = width * chartMarginPercent;
		double heightMargin = height * chartMarginPercent;

		double chartY = 0;
		double widthRemaining = width;
		double heightRemaining = height;

		if(style.isShowTitle()) {
			titleControl.setBounds(0, 0, width - 1, heightMargin - 1);
			heightRemaining -= heightMargin;
			chartY = heightMargin;
		}
		if(style.isShowLegend()) {
			if(style.getLegendStyle().isVerticalLayout()) {
				legendControl.setBounds(width - widthMargin, height - heightRemaining,
						widthMargin - 1, heightRemaining - 1);
				widthRemaining -= widthMargin;
			} else {
				legendControl.setBounds(0, height - heightMargin, width - 1, heightMargin - 1);
				heightRemaining -= heightMargin;
			}
		}
		chartControl.setBounds(0, chartY, widthRemaining, heightRemaining);
	}

	/** Draws the spider chart. */
	private void drawChart(GraphicsContext gc) {
		if(style.isShowTitle()) {
			titleControl.draw(gc);
		}
		chartControl.draw(gc);
		if(style.isShowLegend()) {
			legendControl.draw(gc);
		}
	}
}
