/*******************************************************************************
 * Copyright (c) 2017, 2018 fortiss GmbH. 
 * 
 * This program and the accompanying materials are made available under the
 * terms of the Apache License, Version 2.0, which is available at 
 * https://www.apache.org/licenses/LICENSE-2.0.
 *
 * SPDX-License-Identifier: Apache-2.0
 *******************************************************************************/
package org.fortiss.tooling.spiderchart.style;

import static javafx.scene.paint.Color.LIGHTGRAY;
import static org.fortiss.tooling.common.ui.javafx.style.FontStyle.BLACK_VERDANA_10PT;
import static org.fortiss.tooling.common.ui.javafx.style.LineStyle.SOLID_BLACK_1PT;

import java.text.DecimalFormat;

import org.fortiss.tooling.common.ui.javafx.style.FillStyle;
import org.fortiss.tooling.common.ui.javafx.style.FontStyle;
import org.fortiss.tooling.common.ui.javafx.style.LineStyle;
import org.fortiss.tooling.spiderchart.model.AxisBase;
import org.fortiss.tooling.spiderchart.model.DataSeries;
import org.fortiss.tooling.spiderchart.model.DoubleAxis;
import org.fortiss.tooling.spiderchart.model.EnumerationAxis;

/**
 * Class for all style attributes of {@link DataSeries data series}.
 * 
 * @author hoelzl
 */
public final class DataSeriesStyle {
	/** The line style of the data series. */
	private final LineStyle lineStyle;
	/** The fill style of the data series. */
	private final FillStyle fillStyle;
	/** The font style of the data series indicators. */
	private final FontStyle indicatorLabelStyle;
	/** The size of the indicator box. Default is 7. */
	private final double indicatorSize;
	/** The flag to show indicator boxes. */
	private final boolean showIndicators;
	/** The flag to show indicator labels. */
	private final boolean showIndicatorLabels;
	/** The decimal formatter of the segment labels of the axis. */
	private DecimalFormat decimalFormat;

	/** Constructor. */
	public DataSeriesStyle(LineStyle lineStyle, FillStyle fillStyle, boolean withIndicators,
			boolean withIndicatorLabels, FontStyle indicatorStyle, double indicatorSize,
			DecimalFormat decimalFormat) {
		this.lineStyle = lineStyle;
		this.fillStyle = fillStyle;
		this.showIndicators = withIndicators;
		this.showIndicatorLabels = withIndicatorLabels;
		this.indicatorLabelStyle = indicatorStyle;
		this.indicatorSize = indicatorSize;
		this.decimalFormat = decimalFormat;
	}

	/** Constructor. */
	public DataSeriesStyle() {
		this(SOLID_BLACK_1PT, new FillStyle(LIGHTGRAY), true, true, BLACK_VERDANA_10PT, 7,
				new DecimalFormat("#.##"));
	}

	/** Returns line style. */
	public LineStyle getLineStyle() {
		return lineStyle;
	}

	/** Returns fill style. */
	public FillStyle getFillStyle() {
		return fillStyle;
	}

	/** Returns whether to show indicators. */
	public boolean isShowIndicators() {
		return showIndicators;
	}

	/** Returns whether to show indicator labels. */
	public boolean isShowIndicatorLabels() {
		return showIndicatorLabels;
	}

	/** Returns the indicator label style. */
	public FontStyle getIndicatorLabelStyle() {
		return indicatorLabelStyle;
	}

	/** Returns the size of the indicator box. */
	public double getIndicatorSize() {
		return indicatorSize;
	}

	/** Returns the decimal format. */
	public DecimalFormat getDecimalFormat() {
		return decimalFormat;
	}

	/** Returns the formatted label for the given axis and value. */
	public String getFormattedLabel(AxisBase axis, Object value) {
		if(axis instanceof DoubleAxis && value instanceof Double) {
			return getDecimalFormat().format(value);
		}
		if(axis instanceof EnumerationAxis) {
			return value.toString();
		}
		return null;
	}
}
