/*-------------------------------------------------------------------------+
|                                                                          |
| Copyright 2005-2011 The ConQAT Project                                   |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+-------------------------------------------------------------------------*/
package org.conqat.ide.commons.ui.treemap;

import java.awt.Color;
import java.awt.geom.Rectangle2D;
import java.util.ArrayList;
import java.util.List;

import org.conqat.lib.commons.collections.CollectionUtils;
import org.conqat.lib.commons.string.StringUtils;
import org.conqat.lib.commons.treemap.IDrawingPattern;
import org.conqat.lib.commons.treemap.ITreeMapNode;

/**
 * A treemap node that uses the providers to derive the tree structure.
 * 
 * @author Timo Besenreuther
 * @author $Author: deissenb $
 * @version $Rev: 34252 $
 * @levd.rating GREEN Hash: 949996CAE54F642E0760B5899685F2E7
 */

/* package */class ProviderTreeMapNode<T> implements ITreeMapNode<T> {

	/** Reference to the TreeMap widget */
	private TreeMapWidget<T> widget;

	/** The data object of the TreeMap node */
	private T data;

	/** The layout rectangle */
	private Rectangle2D rect;

	/** The parent node */
	private ProviderTreeMapNode<T> parent;

	/**
	 * The cached children Caching is necessary, because layouter stores info,
	 * the renderer later on reads.
	 */
	private List<ITreeMapNode<T>> children = null;

	/** Constructor */
	public ProviderTreeMapNode(T data, TreeMapWidget<T> widget,
			ProviderTreeMapNode<T> parent) {
		this.data = data;
		this.widget = widget;
		this.parent = parent;
	}

	/** {@inheritDoc} */
	@Override
	public double getArea() {
		if (getChildren().size() > 0) {
			int area = 0;
			for (ITreeMapNode<T> child : getChildren()) {
				area += ((ProviderTreeMapNode<T>) child).getArea();
			}
			return area;
		}
		return widget.getLabelProvider().getArea(data);
	}

	/** {@inheritDoc} */
	@Override
	public List<ITreeMapNode<T>> getChildren() {
		if (children == null) {
			T[] childrenData = widget.getContentProvider().getChildren(data);
			ArrayList<ITreeMapNode<T>> childrenList = new ArrayList<ITreeMapNode<T>>();
			if (childrenData == null) {
				return childrenList;
			}
			for (T childData : childrenData) {
				childrenList.add(new ProviderTreeMapNode<T>(childData, widget,
						this));
			}
			children = CollectionUtils.asUnmodifiable(childrenList);
		}
		return children;
	}

	/** {@inheritDoc} */
	@Override
	public Color getColor() {
		org.eclipse.swt.graphics.Color swtColor = widget.getColorProvider()
				.getColor(data);
		java.awt.Color awtColor = new java.awt.Color(swtColor.getRed(),
				swtColor.getGreen(), swtColor.getBlue());
		return awtColor;
	}

	/** {@inheritDoc} */
	@Override
	public IDrawingPattern getDrawingPattern() {
		IPatternProvider<T> provider = widget.getPatternProvider();
		if (provider == null) {
			return null;
		}
		return provider.getPattern(data);
	}

	/** {@inheritDoc} */
	@Override
	public void setLayoutRectangle(Rectangle2D rect) {
		this.rect = rect;
	}

	/** {@inheritDoc} */
	@Override
	public Rectangle2D getLayoutRectangle() {
		return rect;
	}

	/** {@inheritDoc} */
	@Override
	public Color getPatternColor() {
		IPatternProvider<T> provider = widget.getPatternProvider();
		if (provider == null) {
			return null;
		}

		IDrawingPattern pattern = provider.getPattern(data);
		if (pattern instanceof IColoredDrawingPattern) {
			org.eclipse.swt.graphics.Color swtColor = ((IColoredDrawingPattern) pattern)
					.getForegroundColor(data);
			return new Color(swtColor.getRed(), swtColor.getGreen(), swtColor
					.getBlue());
		}
		return Color.GRAY;
	}

	/** {@inheritDoc} */
	@Override
	public String getText() {
		return widget.getLabelProvider().getLabel(data);
	}

	/** {@inheritDoc} */
	@Override
	public T getUserDatum() {
		return data;
	}

	/** Returns the parent node */
	public ProviderTreeMapNode<T> getParent() {
		return parent;
	}

	/**
	 * {@inheritDoc} This implementation is empty, because the method is only
	 * used by the HTML presentation of the TreeMap and not needed in Eclipse.
	 */
	@Override
	public List<String> getTooltipKeys() {
		return CollectionUtils.emptyList();
	}

	/**
	 * {@inheritDoc} This implementation is empty, because the method is only
	 * used by the HTML presentation of the TreeMap and not needed in Eclipse.
	 */
	@Override
	public String getTooltipId() {
		return StringUtils.EMPTY_STRING;
	}

	/**
	 * {@inheritDoc} This implementation is empty, because the method is only
	 * used by the HTML presentation of the TreeMap and not needed in Eclipse.
	 */
	@Override
	public String getTooltipValue(String key) {
		return StringUtils.EMPTY_STRING;
	}

}